# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import os, os.path, re, stat, shutil, struct, urlparse, base64, imghdr

import appy.pod
from appy import commercial
from appy import Object as O
from appy.shared.dav import Resource
from appy.shared import mimeTypesExts
from appy.pod import PodError, getUuid
from appy.shared import utils as sutils
from appy.pod.metadata import MetadataReader
from appy.shared.css import CssStyles, px2cm
from appy.shared.utils import getTempFileName
from appy.pod.odf_parser import OdfEnvironment
from appy.shared.errors import CommercialError

# ------------------------------------------------------------------------------
FILE_NOT_FOUND = "'%s' does not exist or is not a file."
PDF_TO_IMG_ERROR = 'A PDF file could not be converted into images. Please ' \
                   'ensure that Ghostscript (gs) is installed on your ' \
                   'system and the "gs" program is in the path.'
CONVERT_ERROR = 'Program "convert" (imagemagick) must be installed and ' \
  'in the path for performing this operation. If you wanted to convert a ' \
  'SVG file into a PNG file, conversion of SVG files must also be enabled. ' \
  'On Ubuntu: apt-get install librsvg2-bin'
TO_PDF_ERROR = 'ConvertImporter error while converting a doc to PDF: %s.'

# ------------------------------------------------------------------------------
class DocImporter:
    '''Base class used for importing external content into a pod template (an
       image, another pod template, another odt document...'''
    def __init__(self, content, at, format, renderer):
        self.content = content
        self.renderer = renderer
        # If content is None, p_at tells us where to find it (file path, url...)
        self.at = at
        # Check and standardize p_at
        self.format = format
        if at:
            self.at = self.checkAt(at)
        self.res = u''
        self.ns = renderer.currentParser.env.namespaces
        # Unpack some useful namespaces
        self.textNs = self.ns[OdfEnvironment.NS_TEXT]
        self.linkNs = self.ns[OdfEnvironment.NS_XLINK]
        self.drawNs = self.ns[OdfEnvironment.NS_DRAW]
        self.svgNs = self.ns[OdfEnvironment.NS_SVG]
        self.tempFolder = renderer.tempFolder
        self.importFolder = self.getImportFolder()
        # Create the import folder if it does not exist
        if not os.path.exists(self.importFolder): os.mkdir(self.importFolder)
        self.importPath = self.getImportPath()
        # A link to the global fileNames dict (explained in renderer.py)
        self.fileNames = renderer.fileNames
        if self.at:
            # Move the file within the ODT, if it is an image and if this image
            # has not already been imported.
            self.importPath = self.moveFile()
        else:
            # We need to dump the file content (in self.content) in a temp file
            # first. self.content may be binary, a file handler or a
            # FileWrapper.
            if isinstance(self.content, sutils.FileWrapper):
                self.content.dump(self.importPath)
            else:
                if isinstance(self.content, file):
                    fileContent = self.content.read()
                else:
                    fileContent = self.content
                f = file(self.importPath, 'wb')
                f.write(fileContent)
                f.close()
        # Some importers add specific attrs, through method init

    def checkAt(self, at, raiseOnError=True):
        '''Check and apply some transform to p_at'''
        # Resolve relative path
        if at.startswith('./'): at = os.path.join(os.getcwd(), at[2:])
        # Checks that p_at corresponds to an existing file if given
        if raiseOnError and not os.path.isfile(at):
            raise PodError(FILE_NOT_FOUND % at)
        return at

    def getImportFolder(self):
        '''This method gives the path where to dump the content of the document
           or image. In the case of a document it is a temp folder; in the case
           of an image it is a folder within the ODT result.'''
        return '%s/docImports' % self.tempFolder # For most importers

    def getImportPath(self):
        '''Gets the path name of the file to dump on disk (within the ODT for
           images, in a temp folder for docs).'''
        format = self.format
        if not format or format == 'image':
            at = self.at
            if not at or at.startswith('http') or not os.path.exists(at):
                # It is not possible yet to determine the format (will be done
                # later).
                self.format = ''
            else:
                self.format = os.path.splitext(at)[1][1:]
        fileName = '%s.%s' % (getUuid(), self.format)
        return os.path.abspath('%s/%s' % (self.importFolder, fileName))

    def moveFile(self):
        '''In the case "self.at" was used, we may want to move the file at
           self.at within the ODT result in self.importPath (for images) or do
           nothing (for docs). In the latter case, the file to import stays
           at self.at, and is not copied into self.importPath. So the previously
           computed self.importPath is not used at all.'''
        return self.at

class OdtImporter(DocImporter):
    '''This class allows to import the content of another ODT document into a
       pod template.'''

    # POD manages page breaks by inserting a paragraph whose style inserts a
    # page break before or after it.
    pageBreakTemplate = '<text:p text:style-name="podPageBreak%s"></text:p>'
    pageBreaks = O(before=pageBreakTemplate % 'Before',
                   after=pageBreakTemplate % 'After',
                   beforeDuplex=pageBreakTemplate % 'BeforeDuplex')

    def init(self, pageBreakBefore, pageBreakAfter):
        '''OdtImporter-specific constructor'''
        self.pageBreakBefore = pageBreakBefore
        self.pageBreakAfter = pageBreakAfter

    def getPageBreakAfterType(self):
        '''After sub-document insertion, must we insert a page break of type
           "after" or "before"?'''
        if self.pageBreakAfter == 'duplex':
            # Read metadata about the sub-ODT, containing its number of pages
            metadata = MetadataReader(self.importPath).run()
            odd = (metadata.pagecount % 2) == 1
            # Inserting a page break of type "before" will force an additional
            # page break, while inserting a page break of type "after" will only
            # insert a page break if additional content is found after the
            # paragraph onto which the page break style is applied.
            r = odd and 'beforeDuplex' or 'after'
        else:
            r = 'after'
        return r

    def run(self):
        '''Import the content of a sub-ODT document'''
        # Insert a page break before importing the doc if needed
        if self.pageBreakBefore: self.res += self.pageBreaks.before
        # Import the external odt document
        name = getUuid(removeDots=True, prefix='PodSect')
        self.res += '<%s:section %s:name="%s">' \
                    '<%s:section-source %s:href="file://%s" ' \
                    '%s:filter-name="writer8"/></%s:section>' % (
                        self.textNs, self.textNs, name, self.textNs,
                        self.linkNs, self.importPath, self.textNs, self.textNs)
        # Insert (a) page break(s) after importing the doc if needed
        if self.pageBreakAfter:
            type = self.getPageBreakAfterType()
            self.res += getattr(self.pageBreaks, type)
            # Note that if there is no more document content after the last page
            # break, it will not be visible.
        return self.res

class PodImporter(DocImporter):
    '''This class allows to import the result of applying another POD template,
       into the current POD result.'''

    def init(self, context, pageBreakBefore, pageBreakAfter,
             managePageStyles, resolveFields, forceOoCall):
        '''PodImporter-specific constructor'''
        self.context = context
        self.pageBreakBefore = pageBreakBefore
        self.pageBreakAfter = pageBreakAfter
        self.managePageStyles = managePageStyles
        self.resolveFields = resolveFields
        # Force LO call if fields must be resolved, or if we are inserting
        # sub-documents in "duplex" mode. Indeed, in this latter case, we will
        # need to know the exact number of pages for every sub-document. We will
        # call LO for that purpose: it will produce correct document statistics
        # within meta.xml, which is not the case for an ODT document produced by
        # POD without forcing OO call (stats are those from the document
        # template and not from the pod result).
        if (pageBreakAfter == 'duplex') or resolveFields:
            force = True
        else:
            # Inherit from the parent renderer's value or get a specific value
            renderer = self.renderer
            if forceOoCall == renderer.INHERIT:
                force = renderer.forceOoCall
            else:
                # Else, keep the value as passed to p_self's constructor
                force = forceOoCall
        self.forceOoCall = force

    def mustStream(self):
        '''Returns True if communication with LibreOffice must be performed via
           streams.'''
        # The default value is the one from the main renderer
        default = self.renderer.stream
        # Return the default value if no communication with LO must occur
        if not self.forceOoCall: return default
        # Must we communicate with LO via streams ?
        mustStream = (default in (True, 'in')) or \
                     ((default == 'auto') and \
                      (self.renderer.ooServer != 'localhost'))
        # If the global result must be streamed, for this sub-pod, stream must
        # be "in": this way, the output is written on disk, on the distant LO
        # machine. Then, when we will, via the OdtImporter that will be called
        # just after this PodImporter, specify the sub-pod to include, the
        # distant LO will be able to read it from is own disk.
        return mustStream and 'in' or default

    def run(self):
        # This feature is only available in the open source version
        if commercial: raise CommercialError()
        r = self.renderer
        stream = self.mustStream()
        # Define where to store the pod result in the temp folder. If we must
        # stream the result, the temp folder will be on a distant machine, so it
        # must be at the root of the OS temp folder.
        base = (stream == 'in') and sutils.getOsTempFolder() or \
                                    self.getImportFolder()
        resOdt = os.path.join(base, '%s.odt' % getUuid())
        # The POD template is in self.importPath
        template = self.importPath
        # Re-use an existing renderer if we have one that has already generated
        # a result based on this template.
        renderer = r.children.get(template)
        if not renderer:
            renderer = r.clone(template, self.context, resOdt, stream=stream,
              forceOoCall=self.forceOoCall,
              managePageStyles=self.managePageStyles,
              resolveFields=self.resolveFields, metadata=False,
              deleteTempFolder=False)
            renderer.stylesManager.stylesMapping = r.stylesManager.stylesMapping
            # Add this renderer to the list of child renderers on the parent
            r.children[template] = renderer
        else:
            # Recycle this renderer and set him a new result
            renderer.reinit(resOdt, self.context)
        renderer.run()
        # The POD result is in "resOdt". Import it into the main POD result
        # using an OdtImporter.
        odtImporter = OdtImporter(None, resOdt, 'odt', self.renderer)
        odtImporter.init(self.pageBreakBefore, self.pageBreakAfter)
        return odtImporter.run()

class PdfImporter(DocImporter):
    '''This class allows to import the content of a PDF file into a pod
       template. It calls gs to split the PDF into images and calls the
       ImageImporter for importing it into the result.'''
    # Ghostscript devices that can be used for converting PDFs into images
    gsDevices = {'jpeg': 'jpg', 'jpeggray': 'jpg',
                 'png16m': 'png', 'pnggray': 'png'}

    def run(self):
        # This feature is only available in the open source version
        if commercial: raise CommercialError()
        imagePrefix = os.path.splitext(os.path.basename(self.importPath))[0]
        # Split the PDF into images with Ghostscript. Create a sub-folder in the
        # OS temp folder to store those images.
        imagesFolder = sutils.getOsTempFolder(sub=True)
        device = 'png16m'
        ext = PdfImporter.gsDevices[device]
        dpi = '125'
        cmd = ['gs', '-dSAFER', '-dNOPAUSE', '-dBATCH', '-sDEVICE=%s' % device,
               '-r%s' % dpi, '-dTextAlphaBits=4', '-dGraphicsAlphaBits=4',
               '-sOutputFile=%s/%s%%d.%s' % (imagesFolder, imagePrefix, ext),
               self.importPath]
        sutils.executeCommand(cmd)
        # Check that at least one image was generated
        succeeded = False
        firstImage = '%s1.%s' % (imagePrefix, ext)
        for fileName in os.listdir(imagesFolder):
            if fileName == firstImage:
                succeeded = True
                break
        if not succeeded: raise PodError(PDF_TO_IMG_ERROR)
        # Insert images into the result
        noMoreImages = False
        i = 0
        while not noMoreImages:
            i += 1
            nextImage = '%s/%s%d.%s' % (imagesFolder, imagePrefix, i, ext)
            if os.path.exists(nextImage):
                # Use internally an Image importer for doing this job
                imgImporter = ImageImporter(None, nextImage, ext, self.renderer)
                imgImporter.init('as-char', True, None, None, 'page',
                                 None, True, None)
                self.res += imgImporter.run()
                os.remove(nextImage)
            else:
                noMoreImages = True
        os.rmdir(imagesFolder)
        return self.res

    # Other useful gs commands -------------------------------------------------
    # Convert a PDF from colored to grayscale
    #gs -sOutputFile=grayscale.pdf -sDEVICE=pdfwrite
    #   -sColorConversionStrategy=Gray -dProcessColorModel=/DeviceGray
    #   -dCompatibilityLevel=1.4 -dNOPAUSE -dBATCH colored.pdf
    # Downsample inner images to produce a smaller PDF
    #gs -sOutputFile=smaller.pdf -sDEVICE=pdfwrite
    #   -dCompatibilityLevel=1.4 -dPDFSETTINGS=/screen
    #   -dNOPAUSE -dBATCH some.pdf
    # The "-dPDFSETTINGS=/screen is a shorthand for:
    #-dDownsampleColorImages=true \
    #-dDownsampleGrayImages=true \
    #-dDownsampleMonoImages=true \
    #-dColorImageResolution=72 \
    #-dGrayImageResolution=72 \
    #-dMonoImageResolution=72

class ConvertImporter(DocImporter):
    '''This class allows to import the content of any file that LibreOffice (LO)
       can convert into PDF: doc(x), rtf, xls(x)... It first calls LO to convert
       the document into PDF, then calls a PdfImporter.'''
    def run(self):
        # This feature is only available in the open source version
        if commercial: raise CommercialError()
        # Convert the document into PDF with LibreOffice
        pdfFile = getTempFileName(extension='pdf')
        output = self.renderer.callLibreOffice(self.importPath,
                                               outputName=pdfFile)
        if output: raise PodError(TO_PDF_ERROR % output)
        # Launch a PdfImporter to import this PDF into the POD result
        pdfImporter = PdfImporter(None, pdfFile, 'pdf', self.renderer)
        return pdfImporter.run()

# ------------------------------------------------------------------------------
class Image:
    '''Represents an image on disk. This class is used to detect the image type
       and size.'''
    jpgTypes = ('jpg', 'jpeg')

    def __init__(self, path, format, unit='cm'):
        self.path = path # The image absolute path on disk
        self.format = format or imghdr.what(path)
        # Determine image size in pixels (again, by reading its first bytes)
        self.width, self.height = self.getSizeIn(unit)

    def getSizeIn(self, unit='cm'):
        '''Reads the first bytes from the image on disk to get its size'''
        x = y = None
        # Get the file format from the file name of absent
        format = self.format or os.path.splitext(self.path)[1][1:]
        # Read the file on disk
        f = file(self.path, 'rb')
        if format in Image.jpgTypes:
            # Dummy read to skip header ID
            f.read(2)
            while True:
                # Extract the segment header
                marker, code, length = struct.unpack("!BBH", f.read(4))
                # Verify that it's a valid segment
                if marker != 0xFF:
                    # No JPEG marker
                    break
                elif code >= 0xC0 and code <= 0xC3:
                    # Segments that contain size info
                    y, x = struct.unpack("!xHH", f.read(5))
                    break
                else:
                    # Dummy read to skip over data
                    f.read(length-2)
        elif format == 'png':
            # Dummy read to skip header data
            f.read(12)
            if f.read(4) == "IHDR":
                x, y = struct.unpack("!LL", f.read(8))
        elif format == 'gif':
            imgType = f.read(6)
            buf = f.read(5)
            if len(buf) == 5:
                # else: invalid/corrupted GIF (bad header)
                x, y, u = struct.unpack("<HHB", buf)
        f.close()
        if (x and y) and (unit == 'cm'):
            return float(x)/px2cm, float(y)/px2cm
        else:
            # Return the size, in pixels, or as a tuple (None, None) if
            # uncomputable.
            return x, y

# Compute size of images -------------------------------------------------------
class ImageImporter(DocImporter):
    '''This class allows to import into the ODT result an image stored
       externally.'''
    anchorTypes = ('page', 'paragraph', 'char', 'as-char')
    WRONG_ANCHOR = 'Wrong anchor. Valid values for anchors are: %s.'
    pictFolder = '%sPictures%s' % (os.sep, os.sep)
    # Path of a replacement image, to use when the image to import is not found
    imageNotFound = os.path.join(os.path.dirname(appy.pod.__file__),
                                 'imageNotFound.jpg')

    # Regular expression for finding orientation within EXIF meta-data
    orientationRex = re.compile('exif:Orientation\s*=\s*(\d+)')

    def getZopeImage(self, at):
        '''Gets the Zope Image via an image resolver'''
        resolver = self.renderer.imageResolver
        if not resolver or not at.startswith(resolver.absolute_url()): return
        # The resolver is a Zope application or a sub-object within it. From it,
        # we will retrieve the object on which the image is stored and get the
        # file to download.
        urlParts = urlparse.urlsplit(at)
        path = urlParts[2][1:].split('/')
        image = None
        try:
            image = resolver.unrestrictedTraverse(path)
        except Exception:
            # Maybe a rewrite rule as added some prefix to all URLs ?
            try:
                path = path[1:]
                # Make sure we still have a path, or it gets the resolver
                if path:
                    image = resolver.unrestrictedTraverse(path)
            except Exception:
                pass # The image was not found
        # Return it only if it is an image
        if image and not hasattr(image, 'getBlobWrapper'): image = None
        return image

    def isImage(self, response):
        '''Cheks that the p_response received after a HTTP GET supposed to
           retrieve an image actually contains an image.'''
        if not response or (response.code != 200): return
        # Check p_response's content type
        type = response.headers.get('content-type')
        # The most frequent case: the HTTP GET returns an error page, ie, if the
        # image URL requires authentication.
        if type and not type.startswith('image/'): return
        return True

    def checkAt(self, at):
        '''Do not raise an exception when p_at does not correspond to an
           existing file. We will dump a replacement image instead.'''
        at = DocImporter.checkAt(self, at, raiseOnError=False)
        if at.startswith('http'):
            # Try to get the image
            try:
                response = Resource(at).get(followRedirect=False)
            except (Resource.Error, AttributeError):
                response = None # Can't get the distant image
            if self.isImage(response):
                # Remember the response
                self.httpResponse = response
                return at
            # The HTTP GET did not work, maybe for security reasons (we probably
            # have no permission to get the file). But maybe the URL was a local
            # one, from an application server running this POD code. In this
            # case, if an image resolver has been given to POD, use it to
            # retrieve the image.
            self.zopeImage = self.getZopeImage(at)
            if self.zopeImage:
                return at
            # We could not find the image
            self.format = 'jpg'
            return self.imageNotFound
        elif at.startswith('data:'):
            # The image content and type are embedded in p_at, which comes from
            # the "src" attribute of a HTML "img" tag of the form:
            #              "data:<mimeType>;base64,<base64 content>"
            mimeType = at[5:at.index(';')]
            # The MIME type may be "image/*". In that case, set "img" as virtual
            # format; the real format will be detected once the file will be
            # dumped to disk.
            self.format = mimeTypesExts.get(mimeType) or 'img'
            content = at[at.index(',')+1:]
            # Decode the base64 content and store it in a temp file on disk
            content = base64.b64decode(content)
            fileName = sutils.getTempFileName(extension=self.format)
            f = open(fileName, 'wb')
            f.write(content)
            f.close()
            if self.format == 'img':
                self.format = imghdr.what(fileName)
            return fileName
        else:
            if at.startswith('file://'): at = at[7:]
            if not os.path.isfile(at):
                self.format = 'jpg'
                at = self.imageNotFound
            elif self.format == 'image':
                # Read its format by reading its first bytes
                self.format = imghdr.what(str(at))
            return at

    def getImportFolder(self):
        return os.path.join(self.tempFolder, 'unzip', 'Pictures')

    def moveFile(self):
        '''Copies file at self.at into the ODT file at self.importPath'''
        at = self.at
        importPath = self.importPath
        # Has this image already been imported ?
        for imagePath, imageAt in self.fileNames.iteritems():
            if imageAt == at: # Yes
                i = importPath.rfind(self.pictFolder) + 1
                return importPath[:i] + imagePath
        # The image has not already been imported: copy it
        if not at.startswith('http'):
            shutil.copy(at, importPath)
            # Ensure we can modify the image (with imagemagick)
            os.chmod(importPath, stat.S_IREAD | stat.S_IWRITE)
            return importPath
        # The image has (maybe) been retrieved from a HTTP GET
        response = getattr(self, 'httpResponse', None)
        if response:
            # Retrieve the image format
            format = response.headers['Content-Type']
            if format in mimeTypesExts:
                # At last, I can get the file format
                self.format = mimeTypesExts[format]
                importPath += self.format
                f = file(importPath, 'wb')
                f.write(response.body)
                f.close()
                return importPath
        # The image has (maybe) been retrieved from Zope
        zopeImage = getattr(self, 'zopeImage', None)
        if zopeImage:
            blobWrapper = zopeImage.getBlobWrapper()
            self.format = mimeTypesExts[blobWrapper.content_type]
            importPath += self.format
            blob = blobWrapper.getBlob()
            # If we do not check 'readers', the blob._p_blob_committed is
            # sometimes None.
            blob.readers
            blobPath = blob._p_blob_committed
            shutil.copy(blobPath, importPath)
            return importPath

    def manageExif(self):
        '''Read, with ImageMagick, EXIF metadata in the image, and apply a
           rotation (still with ImageMagick) if required.'''
        # Indeed, LO will not interpret EXIF metadata when reading the image
        cmd = ['identify', '-format', '%[EXIF:*]', self.importPath]
        out, err = sutils.executeCommand(cmd)
        if err: raise Exception(CONVERT_ERROR)
        if out:
            match = self.orientationRex.search(out)
            if match:
                code = int(match.group(1))
                cmd = ['convert', self.importPath,
                       '-auto-orient', self.importPath]
                out, err = sutils.executeCommand(cmd)

    def init(self, anchor, wrapInPara, size, sizeUnit, maxWidth, cssAttrs,
             keepRatio, convertOptions):
        '''ImageImporter-specific constructor'''
        # Initialise anchor
        if anchor not in self.anchorTypes:
            raise PodError(self.WRONG_ANCHOR % str(self.anchorTypes))
        self.anchor = anchor
        self.wrapInPara = wrapInPara
        self.size = size
        self.sizeUnit = sizeUnit
        if maxWidth == 'page':
            maxWidth = self.renderer.stylesManager.pageLayout.getWidth()
        self.maxWidth = maxWidth
        self.keepRatio = keepRatio
        self.convertOptions = convertOptions
        # CSS attributes
        self.cssAttrs = cssAttrs
        if cssAttrs:
            w = getattr(self.cssAttrs, 'width', None)
            h = getattr(self.cssAttrs, 'height', None)
            if w and h:
                self.sizeUnit = w.unit
                self.size = (w.value, h.value)
        # Manage EXIF tags when relevant
        if self.renderer.rotateImages: self.manageExif()
        # Call ImageMagick to perform a custom conversion if required
        options = self.convertOptions
        image = None
        transformed = False
        if options:
            # This feature is only available in the open source version
            if commercial: raise CommercialError()
            if callable(options): # It is a function
                image = Image(self.importPath, self.format)
                options = self.convertOptions(image)
            if options:
                # Ensure we have the right to modify the file@self.importPath
                cmd = ['convert', self.importPath] + options.split() + \
                      [self.importPath]
                out, err = sutils.executeCommand(cmd)
                if err: raise Exception(CONVERT_ERROR)
                transformed = True
        # Avoid creating an Image instance twice if no transformation occurred
        if image and not transformed:
            self.image = image
        else:
            self.image = Image(self.importPath, self.format)

    def getImageSize(self):
        '''Get or compute the image size and returns the corresponding ODF
           attributes specifying image width and height expressed in cm.'''
        # Compute image size, in cm
        width, height = self.image.width, self.image.height
        keepRatio = self.keepRatio
        if self.size:
            # Apply a percentage when self.sizeUnit is 'pc'
            if self.sizeUnit == 'pc':
                # If width or height could not be computed, it is impossible
                if not width or not height: return ''
                if keepRatio:
                    ratioW = ratioH = float(self.size[0]) / 100
                else:
                    ratioW = float(self.size[0]) / 100
                    ratioH = float(self.size[1]) / 100
                width = width * ratioW
                height = height * ratioH
            else:
                # Get, from self.size, required width and height, and convert
                # it to cm when relevant.
                w, h = self.size
                if self.sizeUnit == 'px':
                    w = float(w) / px2cm
                    h = float(h) / px2cm
                # Use (w, h) as is if we don't care about keeping image ratio or
                # if we couldn't determine image's width or height.
                if (not width or not height) or not keepRatio:
                    width, height = w, h
                else:
                    # Compute height and width ratios and apply the minimum
                    # ratio to both width and height.
                    ratio = min(w/width, h/height)
                    width = width * ratio
                    height = height * ratio
        if not width or not height: return ''
        # Take care of max width if specified
        maxWidth = self.maxWidth
        if maxWidth and (width > maxWidth):
            ratio = maxWidth / width
            width = maxWidth
            # Even if keepRatio is False, in that case, nevertheless apply the
            # ratio on heigth, too.
            height = height * ratio
        # Return the ODF attributes
        s = self.svgNs
        return ' %s:width="%fcm" %s:height="%fcm"' % (s, width, s, height)

    def run(self):
        # Some shorcuts for the used xml namespaces
        d = self.drawNs
        t = self.textNs
        x = self.linkNs
        # Compute path to image
        i = self.importPath.rfind(self.pictFolder)
        imagePath = self.importPath[i+1:].replace('\\', '/')
        self.fileNames[imagePath] = self.at
        # In the case of SVG files, perform an image conversion to PNG
        if imagePath.endswith('.svg'):
            newImportPath = os.path.splitext(self.importPath)[0] + '.png'
            out, err = sutils.executeCommand(['convert', self.importPath,
                                             newImportPath])
            if err: raise Exception(CONVERT_ERROR)
            os.remove(self.importPath)
            self.importPath = newImportPath
            imagePath = os.path.splitext(imagePath)[0] + '.png'
            self.format = 'png'
        # Compute image alignment if CSS attr "float" is specified
        floatValue = getattr(self.cssAttrs, 'float', None)
        if floatValue:
            floatValue = floatValue.value.capitalize()
            styleInfo = '%s:style-name="podImage%s" ' % (d, floatValue)
            self.anchor = 'char'
        else:
            styleInfo = ''
        name = getUuid(prefix='I', removeDots=True)
        image = '<%s:frame %s%s:name="%s" %s:z-index="0" ' \
                '%s:anchor-type="%s"%s><%s:image %s:type="simple" ' \
                '%s:show="embed" %s:href="%s" %s:actuate="onLoad"/>' \
                '</%s:frame>' % (d, styleInfo, d, name, d, t, self.anchor,
                self.getImageSize(), d, x, x, x, imagePath, x, d)
        if hasattr(self, 'wrapInPara') and self.wrapInPara:
            style = isinstance(self.wrapInPara, str) and \
                    (' text:style-name="%s"' % self.wrapInPara) or ''
            image = '<%s:p%s>%s</%s:p>' % (t, style, image, t)
        self.res += image
        return self.res
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import hashlib, copy
from appy.px import Px
from appy.utils import asDict
from appy.model.fields import Field
from appy.model.utils import Object as O

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
PARAM_EMPTY = 'Ogone: parameter "%s" has no value and has been ignored.'
RESP_KO     = 'Ogone response SHA failed. Params: %s'
RESP_KO_ERR = 'Failure, fraud? An administrator has been contacted.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Config:
    '''If you plan, in your app, to perform on-line payments via the Ingenico
       ePayments e-Commerce system (previouly named Ogone), create an instance
       of this class in your app and place it in the 'ogone' attr of your app's
       Config class.'''

    # The URL to redirect the user to for performing an on-line payment
    payUrl = 'https://ogone.%s.v-psp.com/ncol/%s/orderstandard_utf8.asp'

    # Fields below are those that must be sent in every request to Ogone
    sendable = ('pspid', 'currency', 'language')

    # Default parameters sent back by Ingenico to your app, after a payment has
    # been performed, being part of the HTTP GET request (payment accepted,
    # refused, etc).
    defaultBackParams = asDict(
      ('ORDERID', 'CURRENCY', 'AMOUNT', 'PM', 'ACCEPTANCE', 'STATUS', 'CARDNO',
       'ED', 'CN', 'TRXDATE', 'PAYID', 'PAYIDSUB', 'NCERROR', 'BRAND', 'IPCTY',
       'CCCTY', 'ECI', 'CVCCHECK', 'AAVCHECK', 'VC', 'DCC_INDICATOR',
       'DCC_EXCHRATE', 'DCC_EXCHRATETS', 'DCC_CONVCCY', 'DCC_CONVAMOUNT',
       'DCC_VALIDHOURS', 'DCC_EXCHRATESOURCE', 'DCC_MARGINPERCENTAGE',
       'DCC_COMMPERCENTAGE', 'IP')
    )

    def __init__(self):
        # self.env refers to the Ogone environment and can be "test" or "prod"
        self.env = 'test'
        # You merchant Ogone ID
        self.pspid = None
        # Default currency for transactions
        self.currency = 'EUR'
        # Default language
        self.language = 'en_US'
        # Key defined by Ingenico, available in your merchant profile on their
        # site, used as element to produce a hash when sending a payment
        # request.
        self.shaInKey = ''
        # Key that you define in your Ingenico merchant profile to check the
        # response coming back from the payment system.
        self.shaOutKey = ''
        # The following parameters must be those you have selected to be
        # included in the URLs called back to your site by Ingenico. You MUST
        # write them all in uppercase letters. If this list contains parameters
        # not being effectively sent by Ingenico, or sent with an empty value,
        # this is not a problem, they will be ignored. But if a single not-empty
        # parameter sent is not this list, the check with the SHA out key will
        # fail.
        # ~~~
        # Do NOT include parameter "SHASIGN" in these parameters
        self.backParams = Config.defaultBackParams

    def __repr__(self): return str(self.__dict__)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Ogone(Field):
    '''This field allows to perform online payments with the Ingenico ePayments
       e-Commerce system.'''

    # Some elements will be traversable
    traverse = Field.traverse.copy()

    urlTypes = ('accept', 'decline', 'exception', 'cancel')

    # Return status codes
    statuses = {
     '0': 'payment_abandoned', # Abandoned or aborted for an unknown reason
     '1': 'payment_canceled',  # Canceled by the customer
     '2': 'payment_refused',   # Payment refused
     '5': 'payment_accepted',  # Payment authorized (in a 2-step process)
     '9': 'payment_accepted',  # Payment accepted
    }

    # The "view" and "cell" layouts are used to display a button for performing
    # an on-line payment.
    # ~~~
    # Note that variable "value" may not be present in the context, because, if
    # the field is shown on layout "buttons", on lists of objects, caller fields
    # like Ref calls directly field.cell instead of field.pxRender.
    view = cell = Px('''
     <!-- The form for sending the payment request to Ogone.
          Variable "value" is misused and contains payment parameters. -->

     <form var="env=config.ogone.env; label=_('pay')"
           method="post" id="form1" name="form1"
           action=":config.ogone.payUrl % (env, env)" class="inline">

       <!-- The list of hidden fields required by Ogone -->
       <input type="hidden" for="key,val in value.items()"
              id=":key" name=":key" value=":val"/>

       <!-- Submit button on most layouts -->
       <input if="layout != 'view'" type="submit" id="submit2" name="submit2"
              value=":label" class="button buttonSmall pay"
              style=":url('pay.svg', bg='18px 18px')"/>

       <!-- Submit picto -->
       <x if="layout == 'view'">
        <input type="image" id="submit2" name="submit2" src=":url('pay.svg')"
               value=":label" class="pictoC" title=":label"/>
        <div align="center">:label</div>
       </x>
     </form>''',

     css='''.pay { border-bottom: 2px solid darkorange !important;
                   font-weight: bold }''')

    edit = search = ''

    def __init__(self, orderMethod, responseMethod, show='view', page='main',
      group=None, layouts=None, move=0, readPermission='read',
      writePermission='write', width=None, height=None, colspan=1, master=None,
      masterValue=None, focus=False, mapping=None, generateLabel=None,
      label=None, view=None, cell=None, edit=None, xml=None, translations=None):
        Field.__init__(self, None, (0,1), None, None, show, page, group,
          layouts, move, False, True, None, None, False, None, readPermission,
          writePermission, width, height, None, colspan, master, masterValue,
          focus, False, mapping, generateLabel, label, None, None, None, None,
          False, False, view, cell, edit, xml, translations)
        # orderMethod must contain a method returning a dict containing info
        # about the order. Following keys are mandatory, among others:
        #   * orderID   An identifier for the order. Don't use the object ID
        #               for this, use a random number, because if the payment
        #               is canceled, Ogone will not allow you to reuse the same
        #               orderID for the next tentative.
        #   * amount    An integer representing the price for this order,
        #               multiplied by 100 (no floating point value, no commas
        #               are tolerated. Dont't forget to multiply the amount by
        #               100!
        # Check Ingenico's doc to determine the complete list of parameters to
        # define. You can use any (upper/lower/camel)case for parameter's names:
        # they will be upperized when sent to Ingenico.
        self.orderMethod = orderMethod
        # responseMethod must contain a method accepting a single parameter. The
        # method will be called when we will get Ogone's response about the
        # status of the payment. The single param is an object whose attributes
        # correspond to all parameters that you have chosen to receive in your
        # Ogone merchant account. Consult their documentation for details. The
        # method must return a tuple (url, message):
        # - url     is the URL to redirect the user to. If None, it will be the
        #           current object's view page. After the payment, the user will
        #           be redirected to that URL.
        # - message is a translated message that will be shown to the user, on
        #           "url". The following standard Appy i18n labels can be used
        #           or overridden to produce this message: "payment_accepted",
        #           "payment_refused", "payment_canceled" and
        #           "payment_abandoned".
        self.responseMethod = responseMethod

    # The Ogone field can a priori be shown on every layout, "buttons" included
    def isRenderable(self, layout): return True

    def createShaDigest(self, values, passphrase):
        '''Creates an Ogone-compliant SHA-512 digest based on key-value pairs in
           dict p_values and on some p_passphrase.'''
        # Create a sorted list of keys
        keys = list(values.keys())
        keys.sort()
        r = []
        for key in keys:
            r.append('%s=%s' % (key, values[key]))
        r = (passphrase.join(r) + passphrase).encode()
        return hashlib.sha512(r).hexdigest()

    def getValue(self, o, name=None, layout=None, single=None):
        '''The "value" of the Ogone field is a dict that collects all the
           necessary info for making the payment.'''
        tool = o.tool
        # Get base parameters from the Ogone config object
        cfg = o.config.ogone
        r = O()
        for name in cfg.sendable:
            setattr(r, name.upper(), getattr(cfg, name))
        inKey = cfg.shaInKey
        # Add dynamic parameters (order ID, amount, client information...)
        for name, value in self.callMethod(o, self.orderMethod).items():
            # Ignore any empty parameter
            if value is None: 
                o.log(PARAM_EMPTY % name)
                continue
            # Convert the value to a str if it is not the case
            if not isinstance(value, str):
                value = str(value).strip()
            # Ignore any empty parameter
            if not value: 
                o.log(PARAM_EMPTY % name)
                continue
            # Store the parameter
            setattr(r, name.upper(), value)
        # Add standard back URLs
        r.CATALOGURL = r.HOMEURL = o.siteUrl
        # Add redirect URLs
        for t in self.urlTypes:
            setattr(r, ('%surl' % t).upper(),
                    '%s/%s/process' % (o.url, self.name))
        # Compute a SHA-512 key as required by Ogone and add it to the result
        r.SHASIGN = self.createShaDigest(r, inKey)
        return r

    def ogoneResponseOk(self, ogoneSign, outKey, params):
        '''Returns True if the SHA-512 signature from Ogone matches retrieved
           params.'''
        digest = self.createShaDigest(params, outKey)
        return digest.lower() == ogoneSign.lower()

    def getBackParameters(self, req, configParams):
        '''Extracts, from the request, Ingenico-related parameters'''
        r = O()
        for name, value in req.items():
            # Upperize the name of the parameter
            name = name.upper()
            # Keep it only if mentioned in the config and not empty
            if (name not in configParams) or not value: continue
            setattr(r, name, value)
        return r

    traverse['process'] = 'perm:read'
    def process(self, o):
        '''Processes a response from Ogone'''
        # Get parameters from the request sent back by Ingenico
        cfg = o.config.ogone
        req = o.req
        params = self.getBackParameters(req, cfg.backParams)
        # Call the response method defined in this Ogone field
        if not self.ogoneResponseOk(req.SHASIGN, cfg.shaOutKey, params):
            o.log(RESP_KO % str(o.req.d()))
            raise Exception(RESP_KO_ERR)
        # Call the field method handling the response received from Ogone.
        # Redirect the user to the correct page. If the field method returns
        # some URL, use it. Else, use the p_o's view URL.
        url, message = self.responseMethod(o, params)
        url = url or o.url
        o.H().commit = True
        o.goto(url, message)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from appy.xml import Parser
from appy.xml.escape import Escape

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Cleaner(Parser):
    '''Cleans XHTML content, so it becomes ready to be stored into a
       Appy-compliant format.'''

    # Tags that will never be in the result, content included
    tagsToIgnoreWithContent = ('style', 'head')

    # Tags that will be removed from the result, but whose content will be kept
    tagsToIgnoreKeepContent = ('x', 'html', 'body', 'font', 'center',
                               'blockquote')
    # Attributes to ignore
    attrsToIgnore = ('id', 'name', 'class', 'lang', 'rules')

    # Attrs to add, if not present, to ensure good formatting, be it at the web
    # or ODT levels.
    attrsToAdd = {'table': {'cellspacing':'0', 'cellpadding':'6', 'border':'1'},
                  'tr':    {'valign': 'top'}}

    # Tags that require a line break to be inserted after them
    lineBreakTags = ('p', 'li', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'td')

    # No-end tags
    noEndTags = ('br', 'img')

    def __init__(self, env=None, caller=None, raiseOnError=True,
                 tagsToIgnoreWithContent=tagsToIgnoreWithContent,
                 tagsToIgnoreKeepContent=tagsToIgnoreKeepContent,
                 attrsToIgnore=attrsToIgnore, attrsToAdd=attrsToAdd):
        # Call the base constructor
        Parser.__init__(self, env, caller, raiseOnError)
        self.tagsToIgnoreWithContent = tagsToIgnoreWithContent
        if 'x' not in tagsToIgnoreKeepContent: tagsToIgnoreKeepContent += ('x',)
        self.tagsToIgnoreKeepContent = tagsToIgnoreKeepContent
        self.tagsToIgnore = tagsToIgnoreWithContent + tagsToIgnoreKeepContent
        self.attrsToIgnore = attrsToIgnore
        self.attrsToAdd = attrsToAdd

    def startDocument(self):
        # The result will be cleaned XHTML, joined from self.r
        Parser.startDocument(self)
        self.r = []

    def endDocument(self):
        self.r = ''.join(self.r)

    def startElement(self, tag, attrs):
        e = self.env
        # Dump any previously gathered content if any
        if e.currentContent:
            self.r.append(e.currentContent)
            e.currentContent = ''
        if e.ignoreTag and e.ignoreContent: return
        if tag in self.tagsToIgnore:
            e.ignoreTag = True
            if tag in self.tagsToIgnoreWithContent:
                e.ignoreContent = True
            else:
                e.ignoreContent = False
            e.currentTags.append( (tag, e.ignoreContent) )
            return
        # Add a line break before the start tag if required (ie: xhtml differ
        # needs to get paragraphs and other elements on separate lines).
        if (tag in self.lineBreakTags) and self.r and \
           (self.r[-1][-1] != '\n'):
            prefix = '\n'
        else:
            prefix = ''
        r = '%s<%s' % (prefix, tag)
        # Include the found attributes, excepted those that must be ignored
        for name, value in attrs.items():
            if name in self.attrsToIgnore: continue
            r += ' %s="%s"' % (name, Escape.xml(value))
        # Include additional attributes if required
        if tag in self.attrsToAdd:
            for name, value in self.attrsToAdd[tag].items():
                if name in attrs: continue
                r += ' %s="%s"' % (name, value)
        # Close the tag if it is a no-end tag
        suffix = '/>' if tag in self.noEndTags else '>'
        self.r.append('%s%s' % (r, suffix))

    def endElement(self, tag):
        e = self.env
        if e.ignoreTag and (tag in self.tagsToIgnore) and \
           (tag == e.currentTags[-1][0]):
            # Pop the currently ignored tag
            e.currentTags.pop()
            if e.currentTags:
                # Keep ignoring tags
                e.ignoreContent = e.currentTags[-1][1]
            else:
                # Stop ignoring elems
                e.ignoreTag = e.ignoreContent = False
        elif e.ignoreTag and e.ignoreContent:
            # This is the end of a sub-tag within a region that we must ignore
            pass
        else:
            if self.env.currentContent:
                self.r.append(self.env.currentContent)
            # Close the tag only if it is a no-end tag.
            if tag not in self.noEndTags:
                # Add a line break after the end tag if required (ie: xhtml
                # differ needs to get paragraphs and other elements on separate
                # lines).
                if (tag in self.lineBreakTags) and self.r and \
                   (self.r[-1][-1] != '\n'):
                    suffix = '\n'
                else:
                    suffix = ''
                self.r.append('</%s>%s' % (tag, suffix))
            self.env.currentContent = ''

    def characters(self, content):
        if self.env.ignoreContent: return
        # Remove whitespace
        current = self.env.currentContent
        if not current or (current[-1] == '\n'):
            toAdd = content.lstrip(u'\n\r\t')
        else:
            toAdd = content
        # Re-transform XML special chars to entities
        self.env.currentContent += Escape.xml(toAdd)

    def clean(self, s):
        '''Cleaning XHTML code p_s allows to produce a Appy-compliant,
           ZODB-storable string.'''
        # a. Every <p> or <li> must be on a single line (ending with a carriage
        #    return); else, appy.utils.diff will not be able to compute XHTML
        #    diffs;
        # b. Optimize size: HTML comments are removed
        self.env.currentContent = ''
        # The stack of currently parsed elements (will contain only ignored
        # ones).
        self.env.currentTags = []
        # 'ignoreTag' is True if we must ignore the currently walked tag.
        self.env.ignoreTag = False
        # 'ignoreContent' is True if, within the currently ignored tag, we must
        # also ignore its content.
        self.env.ignoreContent = False
        return self.parse('<x>%s</x>' % s)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
xhtmlInput = '''
<p>
<table class="plain">
<thead>
<tr>
<th class="align-right" align="right">Title column one<br /></th>
<th>title column two</th>
</tr>
</thead>
<tbody>
<tr>
<td class="align-right" align="right">Hi with a <a class="generated" href="http://easi.wallonie.be">http://easi.wallonie.be</a> <br /></td>
<td>fdff</td>
</tr>
<tr>
<td class="align-right" align="right"><br /></td>
<td><br /></td>
</tr>
<tr>
<td class="align-right" align="left">Some text here<br />
<ul><li>Bullet One</li><li>Bullet Two</li>
<ul><li>Sub-bullet A</li><li>Sub-bullet B</li>
<ul><li>Subsubboulette<br /></li></ul>
</ul>
</ul>
</td>
<td>
<table>
<tbody>
<tr>
<td>SubTable</td>
<td>Columns 2<br /></td>
</tr>
</tbody>
</table>
</td>
</tr>
</tbody>
</table>
<br /></p>
'''

xhtmlInput2 = '''
<ul><li>
<p>a</p>
</li><li>
<p>b</p>
</li><li>
<p>c</p>
</li>
<ul>
  <li><p>SUB</p>
  </li>
</ul>
</ul>
'''

xhtmlInput3 = '''
 <table width="500px">
  <tr>
   <td width="876px">AA</td>
   <td width="474px">BB</td>
  </tr>
 </table>
 <table border="1" cellpadding="1" cellspacing="1" style="width:250px">
<tbody>
	<tr>
		<td>1</td>
		<td>Texte texte texte texte texte texte texte texte <s><strong>gras souligné</strong></s> texte texte texte texte texte texte</td>
	</tr>
	<tr>
		<td>2</td>
		<td>Texte texte texte texte texte texte texte texte <em>italique</em> texte texte texte texte texte texte texte</td>
	</tr>
	<tr>
		<td>3</td>
		<td>Texte texte texte texte texte texte texte texte texte <s>barré</s> texte texte texte texte texte texte</td>
	</tr>
</tbody>
</table>'''

xhtmlInput4 = '''
<table>
  <tbody>
    <tr>
      <td colspan="2">SOCIETE</td>
      <td colspan="2">OUTIL</td>
      <td colspan="2">POUBELLES</td>
      <td colspan="2">VELOS</td>
      <td colspan="2">CENDRIERS</td>
      <td colspan="2">TOTAL DE L'OFFRE</td>
    </tr>
  </tbody>
</table>
'''



from appy.pod.test.contexts import Group

groups = [Group('group1'), Group('group2'), Group('toto')]
subListA = [('A1', 'A2', 'A3'), ('AA', 'AB', 'AC'), ('AX', 'AY', 'AZ')]
subListB = [('B1', 'B2'), ('BA', 'BB')]
para='P'



# ------------------------------------------------------------------------------
from appy.gen import Show, Pod, Table
from appy.shared.xml_parser import Escape
from appy.gen.wrappers import AbstractWrapper

# ------------------------------------------------------------------------------
EXPRESSION_ERROR = 'Page "%s" (%s): error while evaluating page expression ' \
                   '"%s" (%s)'

# ------------------------------------------------------------------------------
class PageWrapper(AbstractWrapper):

    def getSubBreadCrumb(self):
        '''Display an icon for downloading this page (and sub-pages if any) as a
           POD.'''
        if not self.isEmpty('parent'): return
        return self.getField('doc').doRender('view', self)

    def showSubPages(self):
        '''For non-writers, show sub-pages only if present'''
        if self.allows('write'): return True
        if not self.isEmpty('pages'): return 'view'

    def validate(self, new, errors):
        '''Inter-field validation.'''
        return self._callCustom('validate', new, errors)

    def showPortlet(self):
        '''Do not show the portlet for a page'''
        return

    def showExpression(self):
        '''Show the expression to managers only'''
        # Do not show it on "view" if empty
        if self.isEmpty('expression'): return Show.V_
        return self.user.hasRole('Manager')

    def mayView(self):
        '''In addition to the workflow, evaluating p_self.expression, if
           defined, determines p_self's visibility.'''
        expression = self.expression
        if not expression: return True
        user = self.user
        try:
            return eval(expression)
        except Exception, err:
            message = EXPRESSION_ERROR % (self.title, self.id,
                                          expression, str(err))
            self.log(message, type='error')
            return True

    def onEdit(self, created):
        return self._callCustom('onEdit', created)

    def getMergedContent(self, level=1):
        '''Returns a chunk of XHTML code containing p_self's info (title and
           content) and, recursively, info about all its sub-pages.'''
        # Add p_self's title
        title = self.getValue('title', formatted='shown')
        r = ['<h%d>%s</h%d>' % (level, Escape.xhtml(title), level)]
        # Add p_self's content
        if not self.isEmpty('content'):
            r.append(self.getValue('content', formatted='shown'))
        # Add sub-pages
        if not self.isEmpty('pages'):
            for page in self.pages:
                r.append(page.getMergedContent(level=level+1))
        return ''.join(r)
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import copy, types, re
from DateTime import DateTime
from persistent.list import PersistentList

from appy.px import Px
from appy import utils
from appy.database import catalog
from appy.tr import FieldTranslations
from appy.utils import string as sutils
from appy.model.utils import Object as O
from appy.model.fields.phase import Page
from appy.ui.layout import Layout, Layouts
from appy.model.fields.group import Group, UiGroup

# In this file, names "list" and "dict" refer to sub-modules. To use Python
# builtin types, use __builtins__['list'] and __builtins__['dict']

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
XSS_DETECTED = 'Detected Javascript in user input.'
XSS_WARNING  = 'Your behaviour is considered a security attack. System ' \
               'administrator has been warned.'
UNINDEXED    = 'Field "%s": cannot retrieve catalog version of unindexed field.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Show:
    '''Provides some frequently used values for field's "show" attributes'''
    # As a convention, a trailing underscore indicates a negation

    # All layouts but edit. To use for users that can consult the field value
    # but cannot modify it. Also used for fields like some Ref fields that are
    # never manipulated via "edit" layouts.
    E_ = ('view', 'result', 'xml')
    # A variant, without "result"
    ER_ = ('view', 'xml')
    # A variant, with layout "buttons" instead of "result". "buttons"
    # corresponds, on lists, to the range of icons and/or buttons present in the
    # "title" column.
    E_B = ('view', 'buttons', 'xml')
    B = ('edit', 'view', 'buttons', 'xml')

    # All layouts but view. To use typically when, on view, the field value is
    # already shown as a part of some custom widget.
    V_ = ('edit', 'result', 'xml')
    # All layouts but view and edit. To use when you need both E_ and V_.
    VE_ = ('result', 'xml')
    # This set is used for showing workflow transitions in lists of objects
    TR = ('view', 'result')
    # This is, a.o., for custom widgets whose edit and view variants are not
    # that different, but that cannot be shown elsewhere (result, xml, etc).
    VE = ('view', 'edit')
    VX = ('view', 'xml')
    EX = ('edit', 'xml')

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Initiator:
    '''From a given field on a given object, it is possible to trigger the
       creation of another object. The most obvious example is the creation of a
       tied object via a Ref field with add=True. An "initiator" represents the
       (object, field) that initiates the object creation.

       In order to customize initiator behaviour, every Field sub-class can
       propose an Initiator sub-class, by overriding its static attribute
       "initiator".'''

    # The concept of initiator can also be used in the context of visualizing
    # objects. For example, when displaying a tied object via a Ref, its
    # "initiator" is considered to be the source object and the source Ref.

    def __init__(self, tool, req, info):
        self.tool = tool
        self.req = req
        # Extract the initiator object and field from p_info, parsed from a
        # "nav" key in the request, or directly containing a tuple (o, field).
        if isinstance(info, str):
            self.info = info.split('.')
            self.o = tool.getObject(self.info[0])
            self.field = self.extractField(req)
        else:
            self.info = ''
            self.o, self.field = info
        # After having created the object, if we are back from a popup, the
        # initiator may force to go back to some URL.
        self.backFromPopupUrl = None

    def extractField(self, req):
        '''Tries to get the initiator field from the request'''
        r = self.o.getField(self.info[1])
        if r is None:
            # This can be an "objectless" initiator. In that case, p_self.o is
            # the tool and the field's class is specified at another index in
            # p_self.info.
            r = self.o.getField(self.info[1], className=self.info[2])
        return r

    def getUrl(self):
        '''Returns the URL for going back to the initiator object, on the page
           showing self.field.'''
        return self.o.getUrl(sub='view', page=self.field.pageName, nav='no')

    def checkAllowed(self, log=False):
        '''Checks that adding an object via this initiator is allowed'''
        return True

    def updateParameters(self, params):
        '''Add the relevant parameters to the object edition page, related to
           this initiator.'''

    def goBack(self):
        '''Once the object has been created, where must we come back ? r_ can be
           - "view"        return to the "view" page of the created object;
           - "initiator"   return to the page, on self.obj, where self.field is.
        '''
        return 'view'

    def getNavInfo(self, new):
        '''If m_goBack is "view" and navigation-related information must be
           shown on the "view" page of the p_new object, this method returns
           it.'''

    def getBackUrl(self, o):
        '''Get the URL to go to after having initiated creation of object p_o'''
        if self.goBack() == 'view':
            # Stay on the "view" page of the newly created object p_o
            r = o.getUrl(nav=self.getNavInfo(o))
        else:
            # Go back to the initiator page
            r = self.getUrl()
        return r

    def manage(self, new):
        '''Once the p_new object has been created, the initiator must surely
           perform a given action with it (ie, for a Ref field, the new object
           must be linked to the initiator object. This is the purpose of this
           method.'''

    def isComplete(self):
        '''We may have thought that an initiator was there, but once parsed, the
           "nav" key may reveal that there is no initiator at all. When it is
           the case, this method returns False and the initiator is considered
           to be inexistent.'''
        return True

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Field:
    '''Basic abstract class for defining any field'''

    # Some elements will be traversable
    traverse = {}

    # Make some related classes available here
    Translations = FieldTranslations
    Layouts = Layouts

    # Some global static variables
    nullValues = (None, '', [], {}, ())
    validatorTypes = (types.FunctionType, type(re.compile('')))
    labelTypes = ('label', 'descr', 'help')
    viewLayouts = ('view', 'cell')
    cellLayouts = ('cell', 'query')

    # Those attributes can be overridden by subclasses for defining,
    # respectively, names of CSS and Javascript files that are required by this
    # field, keyed by layout.
    cssFiles = {}
    jsFiles = {}

    # For most fields, getting its value on some object is automatically done at
    # the framework level. Fields for which this must be disabled override the
    # following static attribute.
    customGetValue = False

    # Some fields are "outer": they are composed or inner fields
    outer = False
    # In most cases, outer fields store persistent lists
    outerType = PersistentList

    # The initiator class related to the Field
    initiator = Initiator

    # The name of the index class storing values of this field in the catalog
    indexType = 'Index' # For most fields, the base Index class suits

    # Render a field. Optional vars:
    # * fieldName   can be given as different as field.name for fields included
    #               in outer fields: in this case, fieldName includes the row
    #               index.
    # * showChanges If True, a variant of the field showing successive changes
    #               made to it is shown.
    # * minimal     If True, the PX related to p_layout is directly called
    #               (view, cell, etc), instead of layout.pxRender. While this
    #               disables some functions like master/slave relationships, it
    #               returns a more lightweight code.
    pxRender = Px('''<x var="minimal=minimal|False;
      showChanges=showChanges|req.showChanges == 'True';
      layout=layout|req.layout;
      isSearch=layout == 'search';
      hostLayout=req.hostLayout;
      name=fieldName or field.name|field.name;
      widgetName='w_%s' % name if isSearch else name;
      rawValue=field.getValueIf(o, name, layout, disable=isSearch);
      value=not isSearch and \
            field.getFormattedValue(o, rawValue, layout, showChanges);
      requestValue=not isSearch and field.getStoredValue(o, name, True);
      inRequest=field.valueIsInRequest(o, req, name, layout);
      error=req['%s_error' % name] or handler.validator.errors.get(name)|None;
      isMultiple=field.isMultiValued();
      masterCss=field.slaves and ('master_%s' % name) or '';
      tagCss=tagCss|None;
      tagCss=field.getTagCss(tagCss, layout);
      o=o or tool;
      tagId='%d_%s' % (o.iid, name);
      tagName=field.master and 'slave' or '';
      layoutTarget=field">:field.getPx(minimal, \
                                       hostLayout or layout, _ctx_)</x>''')

    def doRender(self, layout, o, name=None, minimal=False, tagCss=None,
                 specific=None):
        '''Allows to call pxRender from code, to display the content of this
           field on p_o in some specific context, for example in a Computed
           field.'''
        context = O(layout=layout, field=self, minimal=minimal,
                    name=name or self.name, o=o)
        if tagCss: context.tagCss = tagCss
        if specific: context.update(specific)
        # Get the current PX context, to copy some entries from it. If no such
        # context exists, create one.
        ctx = o.traversal.context or o.traversal.createContext(layout=layout)
        Px.copyContext(context, ctx)
        # Call pxRender
        return self.pxRender(context)

    # Show the field content for some object on a list of referred objects
    pxTied = Px('''
     <!-- The "title" field -->
     <x if="field.name == 'title'">
      <x if="mayView">
       <x if="not ifield.menuUrlMethod or \
              (layout != 'cell')">:ifield.pxObjectTitle</x>
       <x if="ifield.menuUrlMethod and (layout == 'cell')"
          var2="target,url=ifield.getMenuUrl(o, \
           tied)">::ui.Title.get(o, target=target, baseUrl=url)</x>
       <x if="not selector and guard.mayAct(o)">:ifield.pxObjectActions</x>
      </x>
      <div if="not mayView">
       <img src=":url('password.svg')" class="iconS"/>
       <x>:_('unauthorized')</x></div>
     </x>

     <!-- Any other field -->
     <x if="mayView and (field.name != 'title') and \
            field.isShowable(o, 'result')"
        var2="layout='cell';
              fieldName=field.name;
              minimal=not field.inlineEdit">:field.pxRender</x>''')

    # Show the field content for some object on a list of results
    pxResult = Px('''
     <!-- The "title" field -->
     <x if="field.name == 'title'"
        var2="navInfo=mode.getNavInfo(currentNumber); target=mode.target;
              backHook=backHook|None">
      <x if="mayView"
         var2="pageName=o.getDefaultPage('view');
               selectJs=popup and uiSearch.initiator.jsSelectOne(q,cbId) or ''">
       <x if="hasattr(o, 'getSupTitle')">::o.getSupTitle(navInfo)</x>
       <x>::ui.Title.get(o, mode=mode.titleMode, nav=navInfo, target=target, \
          page=pageName, popup=popup, selectJs=selectJs, highlight=True, \
          backHook=backHook, pageLayout=uiSearch.search.pageLayoutOnView)</x>
       <span if="hasattr(o,'getSubTitle')"
             class=":class_.getCssFor(o,'subTitle')"
             style=":'display:inline' if mode.showSubTitles else 'display:none'"
             name="subTitle">::uiSearch.highlight(o.getSubTitle())</span>

       <!-- Actions -->
       <div if="not popup and uiSearch.showActions and guard.mayAct(o)"
            class=":ui.getActionsCss(uiSearch.actionsDisplay)"
            var2="layout='buttons';
                  editable=guard.mayEdit(o);
                  locked=o.Lock.isSet(o, user, 'main')">

        <!-- Fields on layout "buttons" -->
        <x if="not popup and (uiSearch.showActions == 'all')"
           var2="fields=o.getFields('buttons', 'main'); layout='cell'">
         <!-- Call cell and not pxRender to avoid having a table -->
         <x for="field in fields"
            var2="name=field.name;
                  value=field.getValueIf(o, name, layout)">:field.cell</x>
        </x>

        <!-- Transitions -->
        <x if="class_.showTransitions(o, 'result')"
           var2="workflow=o.getWorkflow()">:workflow.pxTransitions</x>

        <!-- Edit -->
        <x if="editable">
         <a if="not locked"
            var2="linkInPopup=popup or (target.target != '_self')"
            target=":target.target"
            onclick=":target.getOnClick(backHook or str(o.iid))"
            href=":o.getUrl(sub='edit', page=o.getDefaultPage('edit'), \
                            nav=navInfo, popup=linkInPopup)">
          <img src=":url('edit.svg')" class="iconS" title=":_('object_edit')"/>
         </a>
         <x if="locked" var2="lockStyle='iconS'; page='main'">::o.Lock.px</x>
        </x>

        <!-- Delete -->
        <img if="not locked and guard.mayDelete(o)" class="clickable iconS"
             src=":url('deleteS.svg')" title=":_('object_delete')"
             onClick=":'onDeleteObject(%s)' % q(o.url)"/>
       </div>
      </x>
      <x if="not mayView">
       <img src=":url('password.svg')" class="iconS"/>
       <x>:_('unauthorized')</x>
      </x>
     </x>
     <!-- Any other field -->
     <x if="mayView and (field.name != 'title') and \
            field.isShowable(o, 'result')"
        var2="layout='cell';
              fieldName=field.name;
              minimal=not field.inlineEdit">:field.pxRender</x>''')

    # Displays a field label
    pxLabel = Px('''<label if="field.hasLabel and field.renderLabel(layout)"
     lfor=":field.name">::_('label', field=field)</label>''')

    # Displays a field description
    pxDescription = Px('''<span if="field.hasDescr"
     class="discreet">::_('descr', field=field)</span>''')

    # Displays a field help (UiGroup has exaclty the same PX)
    pxHelp = UiGroup.pxHelp

    # Displays validation-error-related info about a field
    def validationPx(c):
        '''Python implementation of validation PX'''
        url = c.url('warning.svg')
        if c.error:
            r = '<abbr title="%s"><img src="%s" class="iconS"/></abbr>' % \
                (c.error, url)
        else:
            r = '<img src="%s" class="iconS" style="visibility:hidden"/>' % url
        return r
    
    pxValidation = Px(validationPx)

    # Displays the fact that a field is required
    pxRequired = Px(lambda c: '<span class="required">*</span>')

    # Button for showing changes to the field
    pxChanges = Px('''
     <div if="zobj.hasHistory(name)" style="margin-bottom: 5px">
      <!-- Button for showing the field version containing changes -->
      <input if="not showChanges"
             var2="label=_('changes_show');
                   css=ztool.getButtonCss(label)" type="button" class=":css"
             value=":label" style=":url('changes', bg=True)"
             onclick=":'askField(%s,%s,%s,null,%s)' % \
                       (q(tagId), q(obj.url), q('view'), q('True'))"/>
      <!-- Button for showing the field version without changes -->
      <input if="showChanges"
             var2="label=_('changes_hide');
                  css=ztool.getButtonCss(label)" type="button" class=":css"
             value=":label" style=":url('changesNo', bg=True)"
             onclick=":'askField(%s,%s,%s,null,%s)' % \
                       (q(tagId), q(obj.url), q('view'), q('False'))"/>
     </div>''')

    # Widget for filtering object values on query results
    pxFilterText = Px('''
     <x var="name=field.name;
             filterId='%s_%s' % (mode.hook, name);
             filterIdIcon='%s_icon' % filterId">
      <!-- Pressing the "enter" key in the field clicks the icon (onkeydown) -->
      <input type="text" size="7" id=":filterId"
             value=":mode.filters.get(name, '')"
             onkeydown=":'if (event.keyCode==13) document.getElementById ' \
                         '(%s).click()' % q(filterIdIcon)"/>
      <img id=":filterIdIcon" class="clickable iconS" src=":url('funnel.svg')"
           onclick=":'askBunchFiltered(%s,%s)' % (q(mode.hook), q(name))"/>
     </x>''')

    def __init__(self, validator, multiplicity, default, defaultOnEdit, show,
      page, group, layouts, move, indexed, mustIndex, indexValue,
      emptyIndexValue, searchable, filterField, readPermission, writePermission,
      width, height, maxChars, colspan, master, masterValue, focus, historized,
      mapping, generateLabel, label, sdefault, scolspan, swidth, sheight,
      persist, inlineEdit, view, cell, edit, xml, translations):
        # The validator restricts which values may be defined. It can be an
        # interval (1,None), a list of string values ['choice1', 'choice2'],
        # a regular expression, a custom function, a Selection instance, etc.
        self.validator = validator
        # Multiplicity is a 2-tuple indicating the minimum and maximum
        # occurrences of values.
        self.multiplicity = multiplicity
        # Is the field required or not ? (derived from multiplicity)
        self.required = self.multiplicity[0] > 0
        # Default value
        self.default = default
        # Default value on layout "edit". If None, self.default is used instead.
        self.defaultOnEdit = defaultOnEdit
        # Must the field be visible or not?
        self.show = show
        # When displaying/editing the whole object, on what page and phase must
        # this field value appear ?
        if page:
            self.page = Page.get(page)
            self.pageName = self.page.name
        else:
            self.page = self.pageName = None
        # Within self.page, in what group of fields must this one appear?
        self.group = Group.get(group)
        # The following attribute allows to move a field back to a previous
        # position (useful for moving fields above predefined ones).
        self.move = move
        # If indexed is True, a database index will be set on the field for
        # fast access.
        self.indexed = indexed
        # If "mustIndex", True by default, is specified, it must be a method
        # returning a boolean value. Indexation will only occur when this value
        # is True.
        self.mustIndex = mustIndex
        if not mustIndex and not callable(mustIndex):
            raise Exception('Value for param "mustIndex" must be a method.')
        # For an indexed field, the value stored in the index is deduced from
        # the field value and type. If you want to store, in the index, a value
        # being transformed in some way, specify, in parameter "indexValue" a
        # method, accepting the stored value as single arg, and returning the
        # transformed (or alternative) value. Note that this transform cannot
        # modify the type of data structure. For example, for a Ref, the
        # received data structure is a list of objects; the method must also
        # produce a list of objects.
        self.indexValue = indexValue
        # By default, an object having no value for this field will not be
        # indexed in the corresponding index (provided p_self is indexed). This
        # can be a problem if the index is used for sorting: in that case,
        # because one of the sort algorithms in use walks the index to perform
        # the sort, objects not being present in this index will not appear in
        # the sorted result, although present in the unsorted results. For that
        # purpose, the following attribute may hold, for an object having no
        # value for this field, a special value that will nevertheless be stored
        # in the index. This value must be of the same type as other values for
        # this field; else, comparisons used by the sort algorithms will produce
        # an error. Moreover, this special value can be used in searches. Here
        # are two examples.
        # 1. For a Ref field, suppose "emptyIndexValue" is (0,), which is the
        #    default. Indeed, a valid Ref value is a list or tuple of object
        #    IIDs (=integers). If you want to express a search like the
        #    following:
        # 
        #         or_(tiedObjectId1, tiedObjectId2, << no object at all >>)
        #
        #    you can write it using the special value representing no value:
        #
        #                or_(tiedObjectId1, tiedObjectId2, 0)
        #
        # 2. For a String, suppose emptyIndexValue is "-" (which is the default,
        #    because the "minus" sign is smaller that any digit or letter). If
        #    the field is named "name", you can get all objects having no name
        #    by writing
        #                       name='-'
        #
        # If you are sure you will not use your indexed field for sorting, then,
        # specify "emptyIndexValue" being None. Indeed, it will produce a
        # smaller index: objects with no value for this field will not be
        # present in the index.
        self.emptyIndexValue = emptyIndexValue
        # If specified "searchable", the field will be added to some global
        # index allowing to perform application-wide, keyword searches.
        self.searchable = searchable
        # You may need to specify another field for rendering a filter for
        # p_self. For example, by default in Appy, field "title" uses field
        # "searchable" as filter PX.
        self.filterField = filterField
        # The PX for filtering field values. If None, it means that the field is
        # not filterable. When a p_filterField is specified, its own filter PX
        # will be used for p_self instead of p_self.filterPx.
        self.filterPx = None
        # Normally, permissions to read or write every attribute in a class are
        # granted if the user has the global permission to read or write
        # instances of that class. Those global permissions are represented by
        # strings "read" and "write". If you want a given attribute, or a
        # sub-set of class attributes, to be protected by specific read and/or
        # write permissions, specify, for attributes "readPermission" and/or
        # "writePermission", alternate strings. When defining a workflow for
        # your class, do not forget to define, for every state and for every
        # custom permission you have mentioned in the following attributes, the
        # list of roles that are granted this permission.
        self.readPermission = readPermission
        self.writePermission = writePermission
        # Widget width and height
        self.width = width
        self.height = height
        # While width and height refer to widget dimensions, maxChars hereafter
        # represents the maximum number of chars that a given input field may
        # accept (corresponds to HTML "maxlength" property). "None" means
        # "unlimited".
        self.maxChars = maxChars or ''
        # If the widget is in a group with multiple columns, the following
        # attribute specifies on how many columns to span the widget.
        self.colspan = colspan or 1
        # The list of slaves of this field, if it is a master
        self.slaves = []
        # The behaviour of this field may depend on another, "master" field
        self.setMaster(master, masterValue)
        # If a field must retain attention in a particular way, set focus=True.
        # It will be rendered in a special way.
        self.focus = focus
        # If we must keep track of changes performed on a field, "historized"
        # must be set to True.
        self.historized = historized
        # Mapping is a dict of contexts that, if specified, are given when
        # translating the label, descr or help related to this field.
        self.mapping = self.formatMapping(mapping)
        self.id = id(self)
        self.type = self.__class__.__name__
        self.pythonType = None # The True corresponding Python type
        # Get the layouts. Consult layout.py for more info about layouts.
        self.layouts = layouts = Layouts.getFor(self, layouts)
        # Derive the following attributes from the layouts, determining the
        # presence of the various types of labels on this field.
        self.hasLabel = layouts.hasPart('l')
        self.hasDescr = layouts.hasPart('d')
        self.hasHelp  = layouts.hasPart('h')
        # Can this field have values that can be edited and validated?
        self.validable = True
        # By default, if the base label for a field is in use in at least one of
        # its layouts (in p_self.layouts), it will be generated automatically in
        # your app's .pot and .po files. That being said, if the base label is
        # not used in any layout, but you still want it to be generated, for
        # using it in some other, non standard place like a pod or a custom
        # report, set the following attribute to True.
        self.generateLabel = generateLabel
        # If you want to avoid generating translation labels for this field, and
        # use instead translations already defined on another field, use the
        # following attribute "label". A label is made of 2 parts: the prefix,
        # based on class name, and the name, which is the field name by default.
        # ----------------------------------------------------------------------
        # If "label" is... | it will be understood as...
        # ----------------------------------------------------------------------
        # a string         | a new prefix = a new class name. Your field will
        #                  | use the label of an homonym field on another class.
        # ----------------------------------------------------------------------
        # a tuple          | a tuple (prefix, name), defining both a new prefix
        #                  | and a new field name. Your field will get the label
        #                  | of a field from another class, having another name.
        #                  | And if you want to reuse labels of another field
        #                  | defined on the same class, use a tuple of the form
        #                  |                 (None, name)
        # ----------------------------------------------------------------------
        self.label = label
        # When you specify a default value "for search" (= "sdefault"), on a
        # search screen, in the search field corresponding to this field, this
        # default value will be present.
        self.sdefault = sdefault
        # Colspan for rendering the search widget corresponding to this field.
        self.scolspan = scolspan or 1
        # Width and height for the search widget
        self.swidth = swidth or width
        self.sheight = sheight or height
        # "persist" indicates if field content must be stored in the database.
        # For some fields it is not wanted (ie, fields used only as masters to
        # update slave's selectable values). If you place a method in attribute
        # "persist", similarly, field value will not be stored, but your method
        # will receive it and you will do whatever you want with it.
        self.persist = persist
        # Can the field be inline-edited (on "view" or "cell" layouts)? A method
        # can be specified. If "inlineEdit" is or returns True, a click within
        # the field value as shown on the "view" or "cell" layout will switch
        # the widget to "edit" mode. If it returns 'icon', Switching to "edit"
        # will be done via an "edit" icon.
        self.inlineEdit = inlineEdit
        # If you want to use alternate PXs than Field.view, Field.cell and
        # Field.edit, you can specify it in parameters "view", "cell" and
        # "edit". Instance attributes "view", "cell" and "edit" below will
        # override their corresponding class attributes.
        if view is not None: self.view = view
        if cell is not None: self.cell = cell
        if edit is not None: self.edit = edit
        # Standard marshallers are provided for converting values of this field
        # into XML. If you want to customize the marshalling process, you can
        # define a method in "xml" that will accept a field value and will
        # return a possibly different value. Be careful: do not return a chunk
        # of XML here! Simply return an alternate value, that will be
        # XML-marshalled.
        self.xml = xml
        # The standard system for getting field translations (label,
        # description, etc) is based on translations stored in persistent
        # Translation instances in tool.translations. But one may bypass this
        # system by placing an instance of class
        #            appy.fields.translations.FieldTranslations
        # in this attribute.
        self.translations = translations

    def init(self, class_, name):
        '''Lazy initialisation'''
        # The class into which this field is defined
        self.container = class_
        # The name of the field
        self.name = name
        # Remember master name on every slave
        for slave in self.slaves: slave.masterName = name
        # Determine ids of i18n labels for this field, excepted if translations
        # are already present in self.tranlations.
        if not self.translations:
            prefix = label = None
            if self.label:
                if isinstance(self.label, str):
                    prefix = self.label
                else: # It is a tuple (prefix, name)
                    prefix, label = self.label
            # Complete missing information
            prefix = prefix or class_.name
            label = label or self.name
            # Determine name to use for i18n
            self.labelId = '%s_%s' % (prefix, label)
            self.descrId = '%s_descr' % self.labelId
            self.helpId  = '%s_help' % self.labelId

    def __repr__(self):
        '''String representation for this field'''
        name = '%s::%s' % (self.container.name, self.name) \
               if hasattr(self, 'container') else 'uninit'
        return '<field %s (%s)>' % (name, self.__class__.__name__.lower())

    def isMultiValued(self):
        '''Does this type definition allow to define multiple values?'''
        maxOccurs = self.multiplicity[1]
        return (maxOccurs is None) or (maxOccurs > 1)

    def isSortable(self, inRefs=False):
        '''Can fields of this type be used for sorting purposes, when sorting
           search results (p_inRefs=False) or when sorting Ref fields
           (inRefs=True) ?'''
        if not inRefs:
            return self.indexed and (self.emptyIndexValue is not None) and \
                   not self.isMultiValued()
        # When p_inRef is True, this method will be overridden by fields for
        # which sort is allowed.

    def getFilterPx(self):
        '''Returns the PX to use for rendering a filter for this field'''
        target = self.filterField or self
        name = target.filterPx
        if name:
            r = target, getattr(target, name)
        else:
            r = None, None
        return r

    def isShowable(self, o, layout):
        '''When displaying p_o on a given p_layout, must we show this field ?'''
        # Check if the user has the permission to view or edit the field
        perm = self.writePermission if (layout=='edit') else self.readPermission
        if not o.allows(perm): return
        # Evaluate self.show
        if callable(self.show):
            r = self.callMethod(o, self.show)
        else:
            r = self.show
        # Take into account possible values 'view', 'edit', 'result', etc.
        if type(r) in utils.sequenceTypes:
            return layout in r
        elif r in Layouts.types:
            return r == layout
        # For showing a field on some layouts, they must explicitly be returned
        # by the "show" method.
        if layout not in Layouts.explicitTypes:
            return bool(r)

    def isRenderable(self, layout):
        '''In some contexts, computing m_isShowable can be a performance
           problem. For example, when showing some object's fields on layout
           "buttons", there are plenty of fields that simply can't be shown on
           this kind of layout: it is no worth computing m_isShowable for those
           fields. m_isRenderable is meant to define light conditions to
           determine, before calling m_isShowable, if some field has a chance to
           be shown or not.

           In other words, m_isRenderable defines a "structural" condition,
           independent of any object, while m_isShowable defines a contextual
           condition, depending on some object.'''
        # Most fields are not renderable on layout "buttons"
        return False if layout == 'buttons' else True

    def isClientVisible(self, o):
        '''Returns True if p_self is visible according to master/slave
           relationships.'''
        masterData = self.getMasterData()
        if not masterData: return True
        else:
            master, masterValue = masterData
            if masterValue and callable(masterValue): return True
            # Get the master value from the request
            req = o.req
            if (master.name not in req) and (self.name in req):
                # The master is not there: we cannot say if the slave must be
                # visible or not. But the slave is in the request. So we should
                # not prevent this value from being taken into account.
                return True
            reqValue = master.getRequestValue(o)
            # reqValue can be a list or not
            if type(reqValue) not in utils.sequenceTypes:
                return reqValue in masterValue
            else:
                for m in masterValue:
                    for r in reqValue:
                        if m == r: return True

    def inGrid(self, layout='edit'):
        '''Is this field in a group with style "grid" ?'''
        if not self.group: return
        if isinstance(self.group, __builtins__['dict']):
            for lay, group in self.group.items():
                if lay != layout: continue
                return group and (group.style == 'grid')
            return
        return self.group.style == 'grid'

    def formatMapping(self, mapping):
        '''Creates a dict of mappings, one entry by label type (label, descr,
           help).'''
        if isinstance(mapping, __builtins__['dict']):
            # Is it a dict like {'label':..., 'descr':...}, or is it directly a
            # dict with a mapping?
            for k, v in mapping.items():
                if (k not in self.labelTypes) or isinstance(v, str):
                    # It is already a mapping
                    return {'label':mapping, 'descr':mapping, 'help':mapping}
            # If we are here, we have {'label':..., 'descr':...}. Complete
            # it if necessary.
            for labelType in self.labelTypes:
                if labelType not in mapping:
                    mapping[labelType] = None # No mapping for this value
            return mapping
        else:
            # Mapping is a method that must be applied to any i18n message
            return {'label':mapping, 'descr':mapping, 'help':mapping}

    def getPx(self, minimal, layout, context):
        '''Returns the PX corresponding to p_layout'''
        if minimal:
            # Call directly the layout-related PX on the field (bypass the
            # layout PX)
            r = getattr(self, layout)
        else:
            # Get the Layout instance related to p_layout and render its PX
            if layout in Field.cellLayouts:
                table = self.Layouts.cell
            else:
                table = self.layouts[layout]
            # Add the layout in the context, it is required by its PX
            context['table'] = table
            r = table.pxRender
        return r

    def setMandatoriness(self, required):
        '''Updating mandatoriness for a field is not easy, this is why this
           method exists.'''
        # Update attributes "required" and "multiplicity"
        self.required = required
        self.multiplicity = (1 if required else 0), self.multiplicity[1]
        # Update the "edit" layout. Clone the layouts: they may be shared by
        # other fields.
        layouts = self.layouts.clone()
        layouts['edit'].setRequired(required)
        self.layouts = layouts

    def setMaster(self, master, masterValue):
        '''Initialises the master and the master value if any'''
        self.master = master
        if master: self.master.slaves.append(self)
        # The semantics of attribute "masterValue" below is as follows:
        # - if "masterValue" is anything but a method, the field will be shown
        #   only when the master has this value, or one of it if multivalued;
        # - if "masterValue" is a method, the value(s) of the slave field will
        #   be returned by this method, depending on the master value(s) that
        #   are given to it, as its unique parameter.
        self.masterValue = utils.initMasterValue(masterValue)

    def getCss(self, layout, r):
        '''Complete list p_r with the names of CSS files that are required for
           displaying widgets of self's type on a given p_layout. p_r is not a
           set because order of inclusion of CSS files may be important.'''
        if layout in self.cssFiles:
            for name in self.cssFiles[layout]:
                if name not in r:
                    r.append(name)

    def getJs(self, layout, r, config):
        '''Completes list p_r with the names of Javascript files that are
           required for displaying widgets of self's type on a given p_layout.
           p_r is not a set because order of inclusion of Javascript files may
           be important.'''
        if layout in self.jsFiles:
            for name in self.jsFiles[layout]:
                if name not in r:
                    r.append(name)

    def isInner(self):
        '''Returns True if p_self is an inner field within a container field'''
        return '*' in self.name

    def getStoredValue(self, o, name=None, fromRequest=False):
        '''Gets the value in its form as stored in the database, or in the
           request if p_fromRequest is True. It differs from calling
           m_getRequestValue because here, in the case of an inner field, the
           request value is searched within the outer value build and stored on
           the request.'''
        if self.isInner():
            # p_self is a sub-field into an outer field: p_name is of the form
            #            [outerName]*[name]*[rowId]
            outerName, name, rowId = name.split('*')
            if rowId == '-1': r = None # It is a template row
            else:
                # Get the outer value
                outerField = o.getField(outerName)
                r = o.req[outerName] if fromRequest else outerField.getValue(o)
                # Access the inner value
                if r:
                    if outerField.outerType == PersistentList:
                        # The row ID is the index if an element from a list
                        rowId = int(rowId)
                        if rowId < len(r):
                            r = getattr(r[rowId], name, None)
                        else:
                            r = None
                    else:
                        # The row ID is a key from a dict
                        r = r.get(rowId, None)
                        if r: r = r.get(name, None)
            # Return an empty string if fromRequest is True
            if fromRequest and (r is None): r = ''
        else:
            r = self.getRequestValue(o, self.name) if fromRequest \
                else o.values.get(self.name)
        return r

    def getRequestValue(self, o, requestName=None):
        '''Gets a value for this field as carried in the request. In the
           simplest cases, the request value is a single value whose name in the
           request is the name of the field.

           Sometimes, several request values must be combined (ie: see the
           overriden method in the Date class).

           Sometimes (ie, a field within a List/Dict), the name of the request
           value(s) representing the field value does not correspond to the
           field name (ie: the request name includes information about
           the container field). In this case, p_requestName must be used for
           searching into the request, instead of the field name (self.name).'''
        return o.req[requestName or self.name]

    def setRequestValue(self, o):
        '''Sets, in the request, field value on p_obj in its "request" form
           (=the way the value is carried in the request).'''
        # Get a copy of the field value on p_obj and put it in the request
        value = self.getCopyValue(o)
        if value is not None:
            o.req[self.name] = value

    def getRequestSuffix(self):
        '''In most cases, in the user interface, there is one homonym HTML
           element for representing this field. Sometimes, several HTML elements
           are used to represent one field (ie, for dates: one field for the
           year, one for the month, etc). In this case, every HTML element's
           name has the form <field name><suffix>. This method returns the
           suffix of the "main" HTML element.'''
        return ''

    def getValue(self, o, name=None, layout=None, single=None):
        '''Gets, on p_o(bject), the value for this field (p_self)'''

        # Possible values for parameters are described hereafter.
        # ----------------------------------------------------------------------
        # name    | p_name can be different from self.name if p_self is a
        #         | sub-field into a List: in this case, it includes the row
        #         | number.
        # ----------------------------------------------------------------------
        # layout  | In most cases, we don't care about the layout for getting
        #         | the value of p_self on p_o. One exception is that a default
        #         | value can be defined specifically on layout "edit". In that
        #         | case, p_layout must be available in order to determine the
        #         | possible specific default value for this layout.
        # ----------------------------------------------------------------------
        # single  | This attribute, although defined here at the most abstract
        #         | level, has only sense for "multi-values", like values from
        #         | the Ref field. Such values are always stored as lists or
        #         | tuples but, when field multiplicity is (x, 1), it is
        #         | sometimes convenient to retrieve the single stored value.
        #         | This can be achieved with p_single being True. If p_single
        #         | is None or False, the retrieved value will be the whole
        #         | list or tuple.
        # ----------------------------------------------------------------------
        # Get the value from the database
        value = self.getStoredValue(o, name)
        if self.isEmptyValue(o, value):
            # If there is no value, get the default value if any. Determine
            # which one must be used: p_self.default or p_self.defaultOnEdit.
            if layout == 'edit':
                default = self.defaultOnEdit
                if default is None: default = self.default
            else:
                default = self.default
            # Get the default value, which can be a method
            if callable(default):
                try:
                    # Caching a default value can lead to problems. For example,
                    # the process of creating an object from another one, or
                    # from some data, sometimes consists in (a) creating an
                    # "empty" object, (b) initializing its values and
                    # (c) reindexing it. Default values are computed in (a),
                    # but it they depend on values set at (b), and are cached
                    # and indexed, (c) will get the wrong, cached value.
                    default = self.callMethod(o, default, cache=False)
                except Exception:
                    # Already logged. Here I do not raise the exception,
                    # because it can be raised as the result of reindexing
                    # the object in situations that are not foreseen by
                    # method in self.default.
                    default = None
            return default
        return value

    def getValueIf(self, o, name, layout, disable=False):
        '''Special method only called by the framework. For some fields (or
           according to some p_disable condition), value retrieval as performed
           here must be disabled.'''
        if disable or self.customGetValue: return
        return self.getValue(o, name, layout)

    def getCopyValue(self, o):
        '''Gets the value of this field on p_o as with m_getValue above. But
           if this value is mutable, get a copy of it.'''
        return self.getValue(o)

    def getComparableValue(self, o):
        '''Get the value of this field on p_o, as can be compared with its new
           version, coming from the validator.'''
        # For immutable values, it simply returns the currently stored value.
        # For mutable values, there are 2 cases:
        # (a) the value as produced by the validator completely overwrites the
        #     existing value. Thus, the previous and new versions can safely be
        #     compared, because being completely distinct. In this case, as for
        #     immutable values, the currently stored value can also be used
        #     as-is, as "comparable" value ;
        # (b) the value as produced by the validator updates the stored value.
        #     In this case, comparing the previous and new versions would lead
        #     to comparing the same value. Here, the "comparable" value must be
        #     a deep copy of the previous value and not the stored value itself.
        return self.getValue(o, single=False)

    def getFormattedValue(self, o, value, layout='view', showChanges=False,
                          language=None):
        '''p_value is a real p_o(bject) value from a field from this type. This
           method returns a pretty, string-formatted version, for displaying
           purposes. Needs to be overridden by some child classes. If
           p_showChanges is True, the result must also include the changes that
           occurred on p_value across the ages. If the formatting implies
           translating some elements, p_language will be used if given, the
           user language else.'''
        if self.isEmptyValue(o, value): return ''
        return value

    def getShownValue(self, o, value, layout='view', showChanges=False,
                      language=None):
        '''Similar to m_getFormattedValue, but in some contexts, only a part of
           p_value must be shown. For example, we may need to display only
           a language-specific part of a multilingual field (see overridden
           method in string.py).'''
        return self.getFormattedValue(o, value, layout, showChanges, language)

    def getXmlValue(self, o, value):
        '''This method allows a developer to customize the value that will be
           marshalled into XML. It makes use of attribute "xml".'''
        return value if not self.xml else self.xml(o, value)

    def updateHistoryValue(self, o, values):
        '''p_values is a dict of field values, of the form
                             ~{s_fieldName: fieldValue}~,
           containing values being stored on p_o prior to a change. Before
           storing this dict in p_o's history, it may be needed to modify, in
           this dict, the (key, value) pair corresponding to p_self.'''
        # In most cases, p_values can be stored as-is, no transform is required

    def getHistoryValue(self, o, value, i, language=None, empty='-'):
        '''Returns p_value as must be shown in the UI within p_o's history. If
           p_language is specified, p_value is the p_language-specific part of a
           multilingual field. p_value is in p_o's history, at entry whose index
           is p_i.'''
        if language is None:
            r = self.getFormattedValue(o, value)
        else:
            r = self.getUniFormattedValue(o, value, contentLanguage=language)
        # If p_r is a list, transform it into a HTML "ul" tag
        if type(r) in utils.sequenceTypes:
            r = '<ul>%s</ul>' % ''.join(['<li>%s</li>' % v for v in r])
        elif not r:
            r = empty
        return r

    def getSearchValue(self, req, widgetName=None, value=None):
        '''Returns the search value (or interval of values) that has been
           encoded in a search form (in p_req) for matching values for this
           field. This value may already be passed in p_value.'''
        if value is not None: return value
        # The base search widget corresponding to p_self is prefixed with "w_".
        # p_widgetName can be specified to get the sub-value of another widget
        # that is part of the complete search widget for p_self.
        widgetName = widgetName or ('w_%s' % self.name)
        # p_req[widgetName] is the base (string) value encoded in the search
        # form. But the search value (or interval of values) can be made of
        # several form fields: overriden methods in Field sub-classes must build
        # search values from these parts, and also convert it to index-compliant
        # values.
        return req[widgetName] or ''

    def searchValueIsEmpty(self, req, widgetName=None):
        '''In some search form, must the search value or interval of values
           specified for this field (p_self) be considered as empty ?'''
        # If, for example, the search widget for this field is made of an
        # interval (from, to), p_widgetName can be given to correspond to the
        # widget part "to", while, by default, the specified part is "from". For
        # some Field sub-classes, the search widget may be even more complex and
        # made of more parts.
        widgetName = widgetName or ('w_%s' % self.name)
        # Conditions in overriden methods can be more complex. For example, if
        # an interval of values (min, max) is expected, specifying only "min" or
        # "max" will allow to perform a search. In this case, we will consider
        # the search value as empty if both "min" and "max" are not specified.
        value = req[widgetName]
        if value and isinstance(value, str): value = value.strip()
        return not value

    def getInlineEditableValue(self, o, value, layout, name=None,
                               language=None):
        '''Returns p_value as it must appear on a view layout, with code
           allowing to inline-edit it when relevant.'''
        # Disable inline-edition if not specified
        inlineEdit = self.getAttribute(o, 'inlineEdit')
        if not inlineEdit: return value
        # Disable inline-edition if the user can't modify the value
        if not o.allows(self.writePermission) or \
           o.Lock.isSet(o, o.user, self.pageName):
            return value
        # Enable inline-edition, either via an additional icon or by clicking
        # in the value.
        suffix = '_%s' % language if language else ''
        hook = '%d_%s%s' % (o.iid, name or self.name, suffix)
        onClick = 'onclick="askField(\'%s\',\'%s\',\'edit:%s\')"' % \
                  (hook, o.url, layout)
        if inlineEdit == 'icon':
            r = '<img src="%s" title="%s" class="inlineIcon iconS" %s/>%s' % \
                (o.buildUrl('edit.svg'), o.translate('object_edit'), onClick,
                 value)
        else:
            r = '<span class="editable" %s>%s</span>' % (onClick, value)
        return r

    def getFakeObject(self, value, req):
        '''Returns a fake object, to mimic an object storing this p_value for
           field p_self.'''
        return O(values={self.name: value}, req=req, id=0, iid=0)

    def getIndexType(self, class_=False):
        '''Returns the name of the class corresponding to the index used to
           store values of this field in a database catalog, or the class itself
           if p_class_ is True.'''
        r = self.indexType
        return eval('catalog.%s' % r) if class_ else r

    def getIndex(self, o):
        '''Gets the Index instance corresponding to this field, or None if the
           field is not indexed.'''
        catalog = o.getCatalog()
        return catalog.get(self.name) if catalog else None

    def getIndexValue(self, o, searchable=False, raw=False):
        '''Return a version for this field value on p_o being ready for indexing
           purposes.'''
        # If p_searchable is True, the value must be a string because it is
        # meant to be included in index "searchable".
        # ~~~
        # If p_raw is True, instead of returning the internal value ready to be
        # indexed, it returns the value before being converted to its internal
        # index representation.
        # ~~~
        # Must we produce an index value ?
        if not self.getAttribute(o, 'mustIndex'): return
        # The indexed value is based on the stored field value
        r = self.getValue(o)
        # Possibly transform the value
        if not searchable:
            r = self.indexValue(o, r) if self.indexValue else r
            # If there is no value to index, get the special value representing
            # an empty value.
            r = self.emptyIndexValue if self.isEmptyValue(o, r) else r
        # Return the value to index (or, if p_raw, the current value in "r")
        if raw: return r
        index = self.getIndexType(True)
        return index.toString(o, r) if searchable else index.toIndexed(r, self)

    def getCatalogValue(self, o, index=None):
        '''Returns the value being currently stored in the catalog for this
           field on p_o.'''
        if not self.indexed: raise Exception(UNINDEXED % self.name)
        # If the corresponding database p_index is not given, get it
        return (index or self.getIndex(o)).getByObject(o, self)

    def hasSortIndex(self):
        '''Some fields have indexes that prevents sorting (ie, list indexes).
           Those fields may define a secondary index, specifically for sorting.
           This is the case of Ref fields for example.'''

    def getSortValue(self, obj):
        '''Return the value of p_self on p_obj that must be used for sorting.
           While the raw p_value may be the value to use in most cases, it is
           not always true. For example, a string like "Gaëtan" could have
           "gaetan" as sort value.'''
        return self.getValue(obj)

    def valueIsInRequest(self, o, req, name=None, layout='view'):
        '''Is there a value corresponding to this field in the request? p_name
           can be different from self.name (ie, if it is a field within another
           (List) field). In most cases, checking that this p_name is in the
           request is sufficient. But in some cases it may be more complex, ie
           for string multilingual fields.'''
        return (name or self.name) in req

    def getStorableValue(self, o, value, complete=False):
        '''p_value is a valid value initially computed through calling
           m_getRequestValue. So, it is a valid representation of the field
           value coming from the request. This method computes the value
           (potentially converted or manipulated in some other way) as can be
           stored in the database.'''
        # More precisely, m_getStorableValue computes a value that can be used
        # as input for m_store. But this latter can further transform it. For
        # example, a value produced by p_getStorableValue for a Ref field can be
        # a list of object IDs; method Ref.store will then convert it to a list
        # of objects and make cross-links between them and p_obj. If p_complete
        # is True, the value produced will be forced to be as close as possible
        # as the database value. In our Ref example, it will be a list of
        # objects (but cross-links will not be established).
        if self.isEmptyValue(o, value): return
        return value

    def getFilterValue(self, value):
        '''p_value is a valid value for p_self, as produced by
           m_getStorableValue. This method applies a potential transform on it
           in order to be used as filter value for a search.'''
        return value

    def getInputValue(self, inRequest, requestValue, value):
        '''Gets the value that must be filled in the "input" widget
           corresponding to this field.'''
        if inRequest:
            return requestValue or ''
        else:
            return value or ''

    def isReadonly(self, o):
        '''Returns True if, when this field is rendered on an edit layout as an
           input field, it must have attribute "readonly" set.'''
        return bool(self.getAttribute(o, 'readonly'))

    def setSlave(self, slaveField, masterValue):
        '''Sets p_slaveField as slave of this field. Normally, master/slave
           relationships are defined when a slave field is defined. At this time
           you specify parameters "master" and "masterValue" for this field and
           that's all. This method is used to add a master/slave relationship
           that was not initially foreseen.'''
        slaveField.master = self
        slaveField.masterValue = utils.initMasterValue(masterValue)
        if slaveField not in self.slaves:
            self.slaves.append(slaveField)
        # Master's init method may not have been called yet
        slaveField.masterName = getattr(self, 'name', None)

    def getMasterData(self):
        '''Gets the master of this field (and masterValue) or, recursively, of
           containing groups when relevant.'''
        if self.master: return self.master, self.masterValue
        group = self.getGroup('edit')
        if group: return group.getMasterData()

    def getMasterTag(self, layout):
        '''Generally, for a field, the name of the tag serving as master for
           driving slave fields is the name of the field itself. But sometimes
           it can be different (see Field sub-classes).'''
        return self.name

    def getTagCss(self, tagCss, layout):
        '''Gets the CSS class(es) that must apply to XHTML tag representing this
           field in the ui. p_tagCss may already give some base class(es).'''
        r = []
        isInner = self.isInner()
        if tagCss and not isInner:
            # For an inner tag, it has no sense to "inherit" from p_tagCss
            r.append(tagCss)
        # Add a special class when this field is the slave of another field
        if self.master:
            css = 'slave*%s*' % self.master.getMasterTag(layout)
            if not callable(self.masterValue):
                css += '*'.join(self.masterValue)
            else:
                css += '+'
            r.insert(0, css)
        # Define a base CSS class when the field is a sub-field in a List
        if isInner: r.append('no')
        return ' '.join(r)

    def getOnChange(self, o, layout, className=None):
        '''When this field is a master, this method computes the call to the
           Javascript function that will be called when its value changes (in
           order to update slaves).'''
        if not self.slaves: return ''
        # When the field is on a search screen, we need p_className
        name = "'%s'" % className if className else 'null'
        return "updateSlaves(this,null,'%s','%s',%s,true)" % \
               (o.url, layout, name)

    def isEmptyValue(self, o, value):
        '''Returns True if the p_value must be considered as an empty value'''
        return value in self.nullValues

    def isCompleteValue(self, o, value):
        '''Returns True if the p_value must be considered as "complete". While,
           in most cases, a "complete" value simply means a "non empty" value
           (see m_isEmptyValue above), in some special cases it is more subtle.
           For example, a multilingual string value is not empty as soon as a
           value is given for some language but will not be considered as
           complete while a value is missing for some language. Another example:
           a Date with the "hour" part required will not be considered as empty
           if the "day, month, year" part is present but will not be considered
           as complete without the "hour, minute" part.'''
        return not self.isEmptyValue(o, value)

    def validateValue(self, o, value):
        '''This method may be overridden by child classes and will be called at
           the right moment by m_validate defined below for triggering
           type-specific validation. p_value is never empty.'''
        return

    def securityCheck(self, o, value):
        '''This method performs some security checks on the p_value that
           represents user input.'''
        if not isinstance(value, str): return
        # Search Javascript code in the value (prevent XSS attacks)
        if '<script' in value:
            o.log(XSS_DETECTED, type='error')
            raise Exception(XSS_WARNING)

    def validate(self, o, value):
        '''Checks that p_value, coming from the request (p_o is being created or
           edited) and formatted through a call to m_getRequestValue defined
           above, is valid according to this type definition. If it is the case,
           None is returned. Else, a translated error message is returned.'''
        # If the value is required, check that a (complete) value is present
        if not self.isCompleteValue(o, value):
            if self.required and self.isClientVisible(o):
                # If the field is required, but not visible according to
                # master/slave relationships, we consider it not to be required.
                return o.translate('field_required')
            else:
                return
        # Perform security checks on p_value
        self.securityCheck(o, value)
        # Triggers the sub-class-specific validation for this value
        message = self.validateValue(o, value)
        if message: return message
        # Evaluate the custom validator if one has been specified
        value = self.getStorableValue(o, value)
        if self.validator and (type(self.validator) in self.validatorTypes):
            if type(self.validator) != self.validatorTypes[-1]:
                # It is a custom function: execute it
                try:
                    validValue = self.validator(o, value)
                    if isinstance(validValue, str) and validValue:
                        # Validation failed; and p_validValue contains an error
                        # message.
                        return validValue
                    else:
                        if not validValue:
                            return o.translate('field_invalid')
                except Exception as e:
                    return str(e)
                except:
                    return o.translate('field_invalid')
            else:
                # It is a regular expression
                if not self.validator.match(value):
                    return o.translate('field_invalid')

    def store(self, o, value):
        '''Stores, on p_o, the p_value (produced by m_getStorableValue) that
           complies to p_self's type definition.'''
        persist = self.persist
        if not persist: return
        elif callable(persist):
            # Do not store the value but let this method do whatever she wants
            # with it.
            self.persist(o, value)
        else:
            # Store the value on p_o
            o.values[self.name] = value

    def storeValueFromAjax(self, o, value, currentValues):
        '''Called by m_storeValueFromAjax to save p_value on p_o'''
        self.store(o, self.getStorableValue(o, value))
        # Overridden methods may return some custom message part
        return ''

    traverse['storeFromAjax'] = 'perm:write'
    def storeFromAjax(self, o):
        '''Stores the new field value from an Ajax request, or do nothing if
           the action was canceled.'''
        # Be the operation a "save" or "cancel", remove the lock on the page
        # (if is was set).
        o.Lock.remove(o, field=self)
        # Do nothing more when canceling the ajax-save
        req = o.req
        if req.cancel == 'True': return
        # Get the new value
        reqValue = req.fieldContent
        # Are we working on a normal or inner field ?
        if not self.isInner():
            # A normal field. Remember its previous value if it is historized.
            isHistorized = self.getAttribute(o, 'historized')
            currentValues = o.history.getCurrentValues(o, self) \
                            if isHistorized else None
            # Validate the value
            if self.validate(o, reqValue):
                # Be minimalist: do nothing and return the previous value
                return
            # Store the new value on p_o
            part = self.storeValueFromAjax(o, reqValue, currentValues)
            # Update the object history when relevant
            if isHistorized and currentValues:
                o.history.historize(currentValues)
        else:
            # An inner field. No historization (yet) for it.
            part = ''
            # Retrieve the outer, inner fields and row number
            name, px = req.px.rsplit(':', 1)
            outer, inner, i = name.split('*')
            outer = o.getField(outer)
            # Update the value
            value = outer.getValue(o)
            row = value[int(i)]
            setattr(row, inner, reqValue)
            # Store the complete value again on p_o
            o.values[outer.name] = value
        # Update obj's last modification date
        o.modified = DateTime()
        o.reindex()
        o.log('ajax-edited %s%s on %s.' % (self.name, part, o.id))

    def callMethod(self, o, method, cache=True):
        '''This method is used to call a p_method on p_o. p_method is part of
           this type definition (ie a default method, the method of a Computed
           field, a method used for showing or not a field...). Normally, those
           methods are called without any arg. But one may need, within the
           method, to access the related field. This method tries to call
           p_method with no arg *or* with the field arg.'''
        try:
            return o.H().methods.call(o, method, cache=cache)
        except TypeError as te:
            # Try a version of the method that would accept self as an
            # additional parameter. In this case, we do not try to cache the
            # value (we do not call utils.callMethod), because the value may be
            # different depending on the parameter.
            tb = utils.Traceback.get()
            try:
                return method(o, self)
            except Exception as e:
                o.log('method %s:\n%s' % (method.__name__, tb), type='error')
                # Raise the initial error
                raise te

    def getAttribute(self, o, name, cache=True):
        '''Gets the value of attribute p_name on p_self, which can be a simple
           value or the result of a method call on p_o.'''
        r = getattr(self, name)
        return r if not callable(r) else self.callMethod(o, r, cache=cache)

    def getGroup(self, layout):
        '''Gets the group into wich this field is on p_layout'''
        # There may be...
        # (a) ... no group at all
        group = self.group
        if group is None: return
        # (b) ... the same group for all layouts
        elif isinstance(group, Group): return group
        # (c) a specific group for this p_layout
        elif layout in group: return group[layout]

    def getGroups(self):
        '''Returns groups as a list'''
        r = []
        if not self.group: return r
        if isinstance(self.group, Group):
            r.append(self.group)
        else: # A dict
            for group in self.group.values():
                if group and (group not in r):
                    r.append(group)
        return r

    def process(self, obj):
        '''This method is a general hook allowing a field to perform some
           processing after an URL on an object has been called, of the form
           <objUrl>/onProcess.'''
        return obj.goto(obj.absolute_url())

    def renderLabel(self, layoutType):
        '''Indicates if the existing label (if hasLabel is True) must be
           rendered by pxLabel. For example, if field is an action, the
           label will be rendered within the button, not by pxLabel.'''
        if not self.hasLabel: return
        # Label is always shown in search forms
        if layoutType == 'search': return True
        # If the field is within a "tabs" group, the label will already be
        # rendered in the corresponding tab. If the field is in a "grid" group,
        # the label is already rendered in a separate column.
        group = self.getGroup(layoutType)
        if group and (group.style in ('tabs', 'grid')): return
        return True

    def getSelectSize(self, isSearch, isMultiple):
        '''If this field must be rendered as a HTML "select" field, get the
           value of its "size" attribute. p_isSearch is True if the field is
           being rendered on a search screen, while p_isMultiple is True if
           several values must be selected.'''
        if not isMultiple: return 1
        prefix = 's' if isSearch else ''
        height = getattr(self, '%sheight' % prefix)
        if isinstance(height, int): return height
        # "height" can be defined as a string. In this case it is used to define
        # height via a attribute "style", not "size".
        return ''

    def getInputSize(self, sizeAttr=True):
        '''Returns the width of the input field, either in the "size" attribute,
           (p_sizeAttr is True) or as a CSS "width" attribute (p_sizeAttr is
           False).'''
        # If the field width is expressed as a percentage, the width must be
        # specified via the CSS "width" attribute. In all other cases, it is
        # specified via the "size" tag attribute.
        width = self.width
        isPercentage = width and isinstance(width, str) and width.endswith('%')
        if sizeAttr:
            # Tag attribute "size"
            return '' if isPercentage else width
        else:
            # CSS attribute "width"
            return 'width:%s' % width if isPercentage else ''

    def getSelectStyle(self, isSearch, isMultiple):
        '''If this field must be rendered as a HTML "select" field, get the
           value of its "style" attribute. p_isSearch is True if the field is
           being rendered on a search screen, while p_isMultiple is True if
           several values must be selected.'''
        prefix = 's' if isSearch else ''
        # Compute CSS attributes
        res = []
        # Height
        height = getattr(self, '%sheight' % prefix)
        if isMultiple and isinstance(height, str):
            res.append('height: %s' % height)
            # If height is an integer value, it will be dumped in attribute
            # "size", not in CSS attribute "height".
        # Width
        width = getattr(self, '%swidth' % prefix)
        if isinstance(width, str):
            res.append('width: %s' % width)
            # If width is an integer value, it represents a number of chars
            # (usable for truncating the shown values), not a width for the CSS
            # attribute.
        return ';'.join(res)

    def getRadioStyle(self, isSearch=False):
        '''If this field must be rendered as radio buttons or checkboxes, and
           its height is specified as a string, restrict the visible part of the
           widget to this height, with a scrollbar when appropriate.'''
        prefix = 's' if isSearch else ''
        r = []
        height = getattr(self, '%sheight' % prefix)
        if isinstance(height, str):
            r.append('height:%s;overflow-y:auto' % height)
        width = getattr(self, '%swidth' % prefix)
        if isinstance(width, str):
            r.append('width:%s;overflow-x:auto' % width)
        return ';'.join(r)

    def getWidthInChars(self, isSearch):
        '''If attribute "width" contains an integer value, it contains the
           number of chars shown in this field (a kind of "logical" width). If
           it contains a string, we must convert the value expressed in px
           (or in another CSS-compliant unit), to a number of chars.'''
        prefix = isSearch and 's' or ''
        width = getattr(self, '%swidth' % prefix)
        if isinstance(width, int): return width
        if isinstance(width, str) and width.endswith('px'):
            return int(width[:-2]) / 5
        return 30 # Other units are currently not supported

    def getListHeader(self, ctx):
        '''When p_self is used as inner-field, within a table-like rendered
           container field, this method returns the content of the header row
           corresponding to this inner field.'''
        return ctx['_']('label', field=self)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''This module contains classes used for layouting graphical elements
   (fields, widgets, groups, ...).'''

# A layout defines how a given field is rendered in a given context. Several
# contexts exist:
#  "view"   represents a given page for a given Appy class, in read-only mode.
#  "edit"   represents a given page for a given Appy class, in edit mode.
#  "cell"   represents a cell in a table, like when we need to render a field
#           value in a query result or in a reference table.
#  "search" represents an advanced search screen.

# Layout elements for a class or page ------------------------------------------
#  s - The page summary, containing summarized information about the page or
#      class, workflow information and object history.
#  w - The widgets of the current page/class
#  n - The navigation panel (inter-objects navigation)
#  b - The range of buttons (intra-object navigation, save, edit, delete...)

# Layout elements for a field --------------------------------------------------
#  l -  "label"        The field label
#  d -  "description"  The field description (a description is always visible)
#  h -  "help"         Help for the field (typically rendered as an icon,
#                       clicking on it shows a popup with online help
#  v -  "validation"   The icon that is shown when a validation error occurs
#                       (typically only used on "edit" layouts)
#  r -  "required"     The icon that specified that the field is required (if
#                       relevant; typically only used on "edit" layouts)
#  f -  "field"        The field value, or input for entering a value.
#  c -  "changes"      The button for displaying changes to a field

# For every field of a Appy class, you can define, for every layout context,
# what field-related information will appear, and how it will be rendered.
# class "Layouts" below defines the default layouts for pages and fields and
# alternative layouts.
# 
# How to express a layout? You simply define a string that is made of the
# letters corresponding to the field elements you want to render. The order of
# elements correspond to the order into which they will be rendered.

# ------------------------------------------------------------------------------
from appy.px import Px

# ------------------------------------------------------------------------------
rowDelimiters =  {'-':'middle', '=':'top', '_':'bottom'}
rowDelms = ''.join(rowDelimiters.keys())
cellDelimiters = {'|': 'center', ';': 'left', '!': 'right'}
cellDelms = ''.join(cellDelimiters.keys())

pxDict = {
  # Page-related elements
  's': 'pxHeader', 'w': 'pxFields', 'n': 'pxNavigationStrip', 'b': 'pxButtons',
  # Field-related elements
  'l': 'pxLabel', 'd': 'pxDescription', 'h': 'pxHelp', 'v': 'pxValidation',
  'r': 'pxRequired', 'c': 'pxChanges'}

# ------------------------------------------------------------------------------
class Cell:
    '''Represents a cell in a row in a table'''
    def __init__(self, content, align, isHeader=False):
        self.align = align
        self.width = None
        self.content = None
        self.colspan = 1
        if isHeader:
            self.width = content
        else:
            self.content = [] # The list of widgets to render in the cell
            self.decodeContent(content)

    def __repr__(self):
        return '<Cell %s>' % self.content

    def decodeContent(self, content):
        digits = '' # We collect the digits that will give the colspan
        for char in content:
            if char.isdigit():
                digits += char
            else: # It is a letter corresponding to a macro
                self.content.append(pxDict.get(char, char))
        # Manage the colspan
        if digits:
            self.colspan = int(digits)

    def renderContent(self, value, layoutType, layoutTarget):
        '''Renders p_value (one element among self.content) for a given
           p_layoutTarget (a field or object) on some p_layoutType.'''
        if value == 'f':
            # The name of the PX depends on p_layoutType
            return getattr(layoutTarget, 'px%s' % layoutType.capitalize())
        else: # p_value is the name of a PX
            return getattr(layoutTarget, value)

# ------------------------------------------------------------------------------
class Row:
    '''Represents a row in a table'''
    def __init__(self, content, valign, isHeader=False):
        self.valign = valign
        self.cells = []
        self.decodeCells(content, isHeader)
        # Compute the row length
        length = 0
        for cell in self.cells:
            length += cell.colspan
        self.length = length

    def __repr__(self):
        return '<Row %s (%d)>' % (str(self.cells), self.length)

    def decodeCells(self, content, isHeader):
        '''Decodes the given chunk of layout string p_content containing
           column-related information (if p_isHeader is True) or cell content
           (if p_isHeader is False) and produces a list of Cell instances.'''
        cellContent = ''
        for char in content:
            if char in cellDelimiters:
                align = cellDelimiters[char]
                self.cells.append(Cell(cellContent, align, isHeader))
                cellContent = ''
            else:
                cellContent += char
        # Manage the last cell if any
        if cellContent:
            self.cells.append(Cell(cellContent, 'left', isHeader))

# ------------------------------------------------------------------------------
class Table:
    '''Represents a table where to dispose graphical elements'''
    simpleParams = ('style', 'css_class', 'cellpadding', 'cellspacing', 'width',
                    'align')
    derivedRepls = {'view': 'hrvd', 'search': 'hrvd'}

    # Render this Table instance, known in the context as "layout". If the
    # layouted object is a page, the "layout target" (where to look for sub-PXs)
    # will be the object whose page is shown; if the layouted object is a field,
    # the layout target will be this field.

    # Warning: when the layout type is "cell", the cell width and alignment must
    # not be defined by the cell layout, but by the outer column layout.
    pxRender = Px('''
     <table var="layoutCss=layout.css_class;
                 inTd=(layoutType == 'cell') and bool(column)|False"
       cellpadding=":layout.cellpadding"
       cellspacing=":layout.cellspacing"
       width=":not inTd and layout.width or ''"
       align=":not inTd and ztool.flipLanguageDirection(layout.align,dir) or ''"
       class=":tagCss and ('%s %s' % (tagCss, layoutCss)).strip() or layoutCss"
       style=":layout.style" id=":tagId" name=":tagName">
      <!-- The table header row -->
      <tr if="layout.headerRow" valign=":layout.headerRow.valign">
       <th for="cell in layout.headerRow.cells" width=":cell.width"
           align=":ztool.flipLanguageDirection(cell.align, dir)">
       </th>
      </tr>
      <!-- The table content -->
      <tr for="row in layout.rows" valign=":row.valign">
       <td for="cell in row.cells" colspan=":cell.colspan"
           align=":ztool.flipLanguageDirection(cell.align, dir)"
           class=":not loop.cell.last and 'cellGap' or ''">
        <x for="c in cell.content">
         <x>::cell.renderContent(c, layoutType, layoutTarget)</x>
         <img if="not loop.c.last" src=":url('space.gif')"/>
        </x>
       </td>
      </tr>
     </table>''')

    def __init__(self, layoutString=None, style=None, css_class='',
                 cellpadding=0, cellspacing=0, width='100%', align='left',
                 other=None, derivedType=None):
        if other:
            # We need to create a Table instance from another Table instance,
            # given in p_other. In this case, we ignore previous params.
            if derivedType != None:
                # We will not simply mimic p_other. If p_derivedType is:
                # - "view", p_other is an "edit" layout, and we must create the
                #           corresponding "view" layout;
                # - "search", p_derivedFrom is a "view" layout.
                self.layoutString = Table.deriveLayout(other.layoutString,
                                                       derivedType)
            else:
                self.layoutString = other.layoutString
            source = 'other.'
        else:
            source = ''
            self.layoutString = layoutString
        # Initialise simple params, either from the true params, either from
        # the p_other Table instance.
        for param in Table.simpleParams:
            exec 'self.%s = %s%s' % (param, source, param)
        # The following attribute will store a special Row instance used for
        # defining column properties.
        self.headerRow = None
        # The content rows will be stored hereafter.
        self.rows = []
        self.decodeRows(self.layoutString)

    @staticmethod
    def deriveLayout(layout, derivedType):
        '''Returns a layout derived from p_layout'''
        res = layout
        for letter in Table.derivedRepls[derivedType]:
            res = res.replace(letter, '')
        # Strip the derived layout
        res = res.lstrip(rowDelms); res = res.lstrip(cellDelms)
        return res

    def addCssClasses(self, css_class):
        '''Adds a single or a group of p_css_class.'''
        if not self.css_class: self.css_class = css_class
        else:
            self.css_class += ' ' + css_class
            # Ensures that every class appears once
            self.css_class = ' '.join(set(self.css_class.split()))

    def isHeaderRow(self, rowContent):
        '''Determines if p_rowContent specified the table header row or a
           content row.'''
        # Find the first char that is a number or a letter
        for char in rowContent:
            if char not in cellDelimiters:
                if char.isdigit(): return True
                else:              return False
        return True

    def decodeRows(self, layoutString):
        '''Decodes the given p_layoutString and produces a list of Row
           instances.'''
        # Split the p_layoutString with the row delimiters
        rowContent = ''
        for char in layoutString:
            if char in rowDelimiters:
                valign = rowDelimiters[char]
                if self.isHeaderRow(rowContent):
                    if not self.headerRow:
                        self.headerRow = Row(rowContent, valign, isHeader=True)
                else:
                    self.rows.append(Row(rowContent, valign))
                rowContent = ''
            else:
                rowContent += char
        # Manage the last row if any
        if rowContent:
            self.rows.append(Row(rowContent, 'middle'))

    def removeElement(self, elem):
        '''Removes given p_elem from myself'''
        macroToRemove = pxDict[elem]
        for row in self.rows:
            for cell in row.cells:
                if macroToRemove in cell.content:
                    cell.content.remove(macroToRemove)
        if elem in self.layoutString:
            self.layoutString = self.layoutString.replace(elem, '')

    def __repr__(self): return '<Table %s>' % self.layoutString

# ------------------------------------------------------------------------------
class ColumnLayout:
    '''A "column layout" dictates the way a table column must be rendered. Such
       a layout is of the form: <name>[*width][,|!|`|`][^].'''
    # --------------------------------------------------------------------------
    #   <name>  | is the name of the field whose content must be shown in
    #           | column's cells ;
    # --------------------------------------------------------------------------
    #   width   | is the width of the column. Any valid value for the "width"
    #           | attribute of the "td" HTML tag is accepted ;
    # --------------------------------------------------------------------------
    #   , | !   | column alignment: respectively, left, centered or right ;
    # --------------------------------------------------------------------------
    #     ^     | if present, indicates that the colmn header must be empty.
    # --------------------------------------------------------------------------

    def __init__(self, layoutString):
        self.layoutString = layoutString

    def get(self):
        '''Returns a list containing the separate elements that are within
           self.layoutString.'''
        consumed = self.layoutString
        lastChar = consumed[-1]
        # Must the column header be empty ?
        if lastChar == '^':
            header = False
            consumed = consumed[:-1]
        else:
            header = True
        # Determine column alignment
        lastChar = consumed[-1]
        if lastChar in cellDelimiters:
            align = cellDelimiters[lastChar]
            consumed = consumed[:-1]
        else:
            align = 'left'
        # Determine name and width
        if '*' in consumed:
            name, width = consumed.rsplit('*', 1)
        else:
            name = consumed
            width = ''
        return name, width, align, header

# ------------------------------------------------------------------------------
class Layouts:
    '''A series of predefined layouts you can use on your Appy fields or pages.
       * First-level static attributes are general layouts applicable to any
         field;
       * default field layouts are in the Field subclass;
       * field-specific layouts are defined on sub-classes whose names
         correspond to Field subclasses;
       * page layouts (the default and others) are defined on the Page
         subclass.'''
    # The following naming conventions apply.
    # "b"  base layout
    # "d"  description (for layouts displaying the field's description)
    # "h"  help (for layouts displaying the field's help icon)
    # "n"  narrow (for non-100%-wide layouts)
    # "w"  wide
    # "g"  group (for layouts applicable to fields in groups with style="grid")
    # "c"  centered
    b = 'lrv-f'
    c = b + '|'
    d = 'lrv-d-f'
    dc = d + '|'
    h = 'lhrv-f'
    n = Table(b, width=None)
    nd = Table(d, width=None)
    w = Table(b, width='100%')
    g = {'edit': 'frvl', 'search': 'l-f'}
    # The *d*escription is visible, even on the *v*iew layout
    dv = {'edit': d, 'view': 'l-d-f'}
    # "Grid" group-related layouts
    gn = {'edit': Table('f;rvl=', width=None), 'search': 'l-f'}
    gd = {'edit': Table('frvl-d', width='99%'), 'search': 'l-f'}
    gdn = Table('d2-f;rvl=', width=None) # "n" for "*n*ot wide"
    gh = 'fhrvl'
    gdh = 'fhrvl-d'

    # ~~~ Default cell layout ~~~
    # Layouts for a field do not contain any entry for layout type "cell".
    # Indeed, rendering a field on such layout is, implicitly, simply rendering
    # its value ("f"). Moreover, rendering a layout renders a complete XHTML
    # table, even if it contains a single cell, like for "f". So we avoid doing
    # this for performance and conciseness reasons. But we can't complelety
    # forget cell layouts, because they are needed for inline-editing (or maybe
    # other Ajax-like functionalities in the future). Indeed, the machinery for
    # ajax-editing the field is inlaid in the XHTML table corresponding to the
    # layout. In such case, we will use a standard cell layout defined here.
    cell = Table('f|', align='center', css_class='no')

    # ~~~ Default layouts for fields ~~~
    class Field:
        defaults = {'normal': {True: 'lrv-f', False: 'lv-f'},
                    'grid':   {True:  {'edit': 'frvl', 'search': 'l-f'},
                               False: {'edit': 'fvl', 'search': 'l-f'}}}

    @classmethod
    def defaults(klass, field):
        '''Returns the default layouts for fields. Depends on:
           - the fact that a field is required or not;
           - the type of group into which the field is.

           If Type-subclass-specific default layouts are defined in
           SubType.getDefaultLayouts, they will override default ones as
           computed here. For every field, the user can also use other
           frequently used layouts that are defined hereabove.'''
        key = field.inGrid() and 'grid' or 'normal'
        return klass.Field.defaults[key][field.required]

    # ~~~ Field-specific layouts ~~~
    class Action:
        b = {'view': 'lf', 'edit': 'lf'}
        c = {'view': 'f|'}

    class Boolean:
        d = {'view': 'lf', 'edit': Table('flrv;=d', width=None)}
        h = {'view': 'lf', 'edit': Table('flhv', width=None)}
        gd = {'view': 'fl', 'edit': Table('f;dv-', width=None)}
        # The "long" version of the previous layout (if the description is
        # long), with vertical alignment on top instead of middle.
        gdl = {'view': 'fl', 'edit': Table('f;dv=', width=None)}
        # Description, renderes as *r*adio buttons
        gdr = {'view': 'fl', 'edit': Table('d-fv=', width=None)}
        rld = {'view': 'lf', 'edit': 'l-d-f'}

    class Calendar:
        b = {'edit': 'f', 'view': 'l-d-f'}
        n = Table('l-f', width=None)

    class Custom:
        # A custom field is not showable by default, except on layout "xml"
        b = 'f'

    class Computed:
        # Wide layout, with content only (no label). Useful for displaying
        # computed warning messages.
        wf = {'view': Table('f', width='100%')}

    class File:
        b = {'view': 'l-f','edit': 'lrv-f'}
        bg = {'view': 'fl','edit': 'frvl'}

    class Info:
        # We specify "edit": that way, "view" inherits from it, and all these
        # layouts are ready in case an info must be shown on an "edit" layout.
        b = {'edit': 'l'}
        d = {'edit': Table('l-d', width=None),
             'view': Table('l-d', width=None)}
        do = {'edit': 'f', 'view': 'd'} # Description only
        c = {'edit': 'l|'} # Centered
        dc = {'edit': 'l|-d|'}
        vdc = {'view': 'l|-d|'}

    class Pod:
        # Right-aligned layouts, convenient for pod fields exporting search
        # results or multi-template pod fields.
        r = {'view': Table('fl!', css_class='podTable')} # "r"ight
        rf = {'view': Table('f!', css_class='podTable')} # "r"ight, no label
        l = {'view': Table('fl;', css_class='podTable')} # "l"eft
        # "r"ight "m"ulti-template (where the global field label is not used
        rm = {'view': Table('f!', css_class='podTable')}

    class Ref:
        # Wide layout for a Ref is a bit different than Layout.d: it must be
        # 100% wide on result, too.
        w = {'view': Table('lrv-f', width='100%')}
        # "d" stands for "description": a description label is added, on view
        wd = {'view': Table('l-d-f', width='100%')}
        # Wide, but not on edit
        w_e = {'edit': Table('lrv-f', width=None),
               'view': Table('lrv-f', width='100%')}
        # "cell" layout varies from the default
        cell = Table('f|', width='100%', css_class='no')

    class String:
        g = {'edit': Table('f;rv=', width=None), 'search': 'l-f'}
        gd = {'edit': Table('f;rv=d', width=None), 'search': 'l-f'}

    class Text:
        b = {'view': 'l-f', 'edit': 'lrv-d-f'}
        g = {'view': Table('fl', width='99%'),
             'edit': Table('d2-f;rv=', width='99%')}

    class Rich:
        b = {'view': 'l-f', 'edit': 'lrv-d-f'}
        c = {'view': 'lc-f', 'edit': 'lrv-d-f'} # Idem but with history
        g = {'view': Table('fl', width='99%'),
             'edit': Table('d2-f;rv=', width='99%')}
        gc = {'view': Table('cl-f', width='99%'),
              'edit': Table('drv-f', width='99%')} # Idem but with history

    class Select:
        d = Table('d2-l-f;rv=', width=None)
        g = {'view': Table('fl', width=None),
             'edit': Table('f;rv=', width=None)}
        gd = {'view': Table('fl', width=None),
              'edit': Table('d2-f;rv=', width=None)}

    class List:
        g = {'view': 'fl', 'edit': Table('f;rv=', width=None), }
        gd = {'view': 'fl', 'edit': Table('d;v=f;', width=None)}

    # ~~~ Layouts for pages ~~~
    class Page:
        # The default layouts
        defaults = {'view': Table('w-b'), 'edit': Table('w-b', width=None)}
        # With the page summary
        summary = {'view': Table('s-w-b'), 'edit': Table('w-b', width=None)}
        # Wide
        wide = {'view': Table('w-b'), 'edit': Table('w-b')}
        # Wide + summary
        wideSummary = {'view': Table('s-w-b'), 'edit': Table('w-b')}
        # Centered
        centered = {'view': Table('w|-b|', align='center'),
                    'edit': Table('w|-b|', width=None, align='center')}
# ------------------------------------------------------------------------------



'''Utility modules and functions for Appy'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from persistent.list import PersistentList

import traceback, mimetypes, subprocess, pdb

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Version:
    '''Appy version'''
    short = '1.0.0'
    verbose = 'Appy %s' % short

    @classmethod
    def get(class_):
        '''Returns a string containing the short and verbose Appy version'''
        if class_.short == 'dev': return 'dev'
        return '%s (%s)' % (class_.short, class_.verbose)

    @classmethod
    def isGreaterThanOrEquals(class_, v):
        '''This method returns True if the current Appy version is greater than
           or equals p_v. p_v must have a format like "0.5.0".'''
        if class_.short == 'dev':
            # We suppose that a developer knows what he is doing
            return True
        else:
            paramVersion = [int(i) for i in v.split('.')]
            currentVersion = [int(i) for i in class_.short.split('.')]
            return currentVersion >= paramVersion

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def asDict(seq):
    '''Returns a dict whose keys are elements from p_seq ad values are None'''
    return {elem: None for elem in seq}

# Global variables - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
sequenceTypes = (list, tuple, PersistentList)
commercial = False

# On these layouts, using a gobal selector and switching from one option to the
# other is not allowed: it would reload the entire page. Examples are: the
# language selector, or the authentication context selector.
noSwitchLayouts = asDict(('edit', 'search'))

# MIME-related stuff - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
od = 'application/vnd.oasis.opendocument'
ms = 'application/vnd.openxmlformats-officedocument'
ms2 = 'application/vnd.ms'

mimeTypes = {'odt':  '%s.text' % od,
             'ods':  '%s.spreadsheet' % od,
             'doc':  'application/msword',
             'docx': '%s.wordprocessingml.document' % ms,
             'rtf':  'text/rtf',
             'pdf':  'application/pdf'
             }

mimeTypesExts = {
    '%s.text' % od:        'odt',
    '%s.spreadsheet' % od: 'ods',
    'application/msword':  'doc',
    'text/rtf':            'rtf',
    'application/pdf':     'pdf',
    'image/png':           'png',
    'image/jpeg':          'jpg',
    'image/pjpeg':         'jpg',
    'image/gif':           'gif',
    '%s.wordprocessingml.document' % ms: 'docx',
    '%s.spreadsheetml.sheet' % ms: 'xlsx',
    '%s.presentationml.presentation' % ms: 'pptx',
    '%s-excel' % ms2:      'xls',
    '%s-powerpoint' % ms2: 'ppt',
    '%s-word.document.macroEnabled.12' % ms2: 'docm',
    '%s-excel.sheet.macroEnabled.12' % ms2: 'xlsm',
    '%s-powerpoint.presentation.macroEnabled.12' % ms2: 'pptm'}

def getMimeType(fileName, default='application/octet-stream'):
    '''Tries to guess mime type from p_fileName'''
    res, encoding = mimetypes.guess_type(fileName)
    if not res:
        if fileName.endswith('.po'):
            res = 'text/plain'
            encoding = 'utf-8'
    if not res: return default
    if not encoding: return res
    return '%s;;charset=%s' % (res, encoding)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class CommercialError(Exception):
    '''Raised when some functionality is called from the commercial version but
       is available only in the free, open source version.'''
    MSG = 'This feature is not available in the commercial version. It is ' \
          'only available in the free, open source (GPL) version of Appy.'
    def __init__(self): Exception.__init__(self, self.MSG)

class MessageException(Exception):
    '''User-friendly exception'''
    # Message exceptions are raised when, due to some problem while handling a
    # browser request, the current traversal must be interrupted, but we don't
    # want a "technical" error to be raised (500). We don't want technical
    # details to be logged nor rendered to the UI, such as a Python traceback:
    # we simply want to display nice, translated info about the problem in a 200
    # response.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class No:
    '''When you write a workflow condition method and you want to return False
       but you want to give to the user some explanations about why a transition
       can't be triggered, do not return False, return an instance of No
       instead. When creating such an instance, you can specify an error
       message.'''
    def __init__(self, msg): self.msg = msg
    def __bool__(self): return False
    def __repr__(self): return '<No: %s>' % self.msg

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def initMasterValue(v):
    '''Standardizes p_v as a list of strings, excepted if p_v is a method'''
    if callable(v): return v
    if not isinstance(v, bool) and not v: res = []
    elif type(v) not in sequenceTypes: res = [v]
    else: res = v
    return [str(v) for v in res]

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def encodeData(data, encoding=None):
    '''Applies some p_encoding to string p_data, but only if an p_encoding is
       specified.'''
    if not encoding: return data
    return data.encode(encoding)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def copyData(data, target, targetMethod, type='string', encoding=None,
             chunkSize=1024):
    '''Copies p_data to a p_target, using p_targetMethod. For example, it copies
       p_data which is a string containing the binary content of a file, to
       p_target, which can be a HTTP connection or a file object.

       p_targetMethod can be "write" (files) or "send" (HTTP connections) or ...
       p_type can be "string" or "file". In the latter case, one may, in
       p_chunkSize, specify the amount of bytes transmitted at a time.

       If an p_encoding is specified, it is applied on p_data before copying.

       Note that if the p_target is a Python file, it must be opened in a way
       that is compatible with the content of p_data, ie open('myFile.doc','wb')
       if content is binary.'''
    dump = getattr(target, targetMethod)
    if not type or (type == 'string'): dump(encodeData(data, encoding))
    elif type == 'file':
        while True:
            chunk = data.read(chunkSize)
            if not chunk: break
            dump(encodeData(chunk, encoding))

# List/dict manipulations  - - - - - - - - - - - - - - - - - - - - - - - - - - -
def splitList(l, sub):
    '''Returns a list that was build from list p_l whose elements were
       re-grouped into sub-lists of p_sub elements.

       For example, if l = [1,2,3,4,5] and sub = 3, the method returns
       [ [1,2,3], [4,5] ].'''
    res = []
    i = -1
    for elem in l:
        i += 1
        if (i % sub) == 0:
            # A new sub-list must be created
            res.append([elem])
        else:
            res[-1].append(elem)
    return res

class IterSub:
    '''Iterator over a list of lists'''
    def __init__(self, l):
        self.l = l
        self.i = 0 # The current index in the main list
        self.j = 0 # The current index in the current sub-list

    def __iter__(self): return self

    def __next__(self):
        # Get the next ith sub-list
        if (self.i + 1) > len(self.l): raise StopIteration
        sub = self.l[self.i]
        if (self.j + 1) > len(sub):
            self.i += 1
            self.j = 0
            return self.__next__()
        else:
            elem = sub[self.j]
            self.j += 1
            return elem

def getElementAt(l, cyclicIndex):
    '''Gets the element within list/tuple p_l that is at index p_cyclicIndex
       (int). If the index out of range, we do not raise IndexError: we continue
       to loop over the list until we reach this index.'''
    return l[cyclicIndex % len(l)]

def flipDict(d, byChar=False):
    '''Flips dict p_d: keys become values, values become keys. p_d is left
       untouched: a new, flipped, dict is returned.'''
    # If p_byChar is True, p_d's values must be strings, and, for every entry
    # of the form
    #                           'k': 'abcd'
    # the following entries are created:
    #               'a': 'k', 'b': 'k', 'c': 'k', 'd': 'k'
    r = {}
    # Browse dict p_d
    for k, v in d.items():
        if byChar:
            for char in v:
                r[char] = k
        else:
            r[v] = k
    return r

def addPair(name, value, d=None):
    '''Adds key-value pair (name, value) to dict p_d. If this dict is None, it
       returns a newly created dict.'''
    if d: d[name] = value
    else: d = {name: value}
    return d

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Traceback:
    '''Dumps the last traceback into a string'''
    @staticmethod
    def get(last=None, html=False):
        '''Gets the traceback as a string. If p_last is given (must be an
           integer value), only the p_last lines of the traceback will be
           included. It can be useful for pod/px tracebacks: when an exception
           occurs while evaluating a complex tree of buffers, most of the
           traceback lines concern uninteresting buffer/action-related recursive
           calls.'''
        r = traceback.format_exc(last)
        if html: r = r.replace('\n', '<br/>')
        return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def executeCommand(cmd, stdin=None):
    '''Executes command p_cmd and returns a tuple (s_stdout, s_stderr)
       containing the data output by the subprocesss on stdout and stderr. p_cmd
       should be a list of args (the 1st arg being the command in itself, the
       remaining args being the parameters), but it can also be a string, too
       (see subprocess.Popen doc).'''
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stdin=stdin,
                            stderr=subprocess.PIPE)
    return proc.communicate()

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
CONVERT_ERROR = 'Program "convert" (ImageMagick) must be installed and ' \
                'in the path for performing this operation (%s).'

def convert(path, options):
    '''Calls the "convert" executable to apply some transform to an image whose
       p_path is passed.'''
    # p_options can be a list, or a string containing blank-separated options
    options = options.split() if isinstance(options, str) else options
    cmd = ['convert', path] + options
    cmd.append(path)
    out, err = executeCommand(cmd)
    if err: raise Exception(CONVERT_ERROR % str(err))
    return out

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def formatNumber(n, sep=',', precision=2, tsep=' ', removeTrailingZeros=False):
    '''Returns a string representation of number p_n, which can be a float
       or integer. p_sep is the decimal separator to use. p_precision is the
       number of digits to keep in the decimal part for producing a nice rounded
       string representation. p_tsep is the "thousands" separator.'''
    if n is None: return ''
    # Manage precision
    if precision is None:
        res = str(n)
    else:
        format = '%%.%df' % precision
        res = format % n
    # Use the correct decimal separator
    res = res.replace('.', sep)
    # Insert p_tsep every 3 chars in the integer part of the number
    splitted = res.split(sep)
    res = ''
    if len(splitted[0]) < 4: res = splitted[0]
    else:
        i = len(splitted[0])-1
        j = 0
        while i >= 0:
            j += 1
            res = splitted[0][i] + res
            if (j % 3) == 0:
                res = tsep + res
            i -= 1
    # Add the decimal part if not 0
    if len(splitted) > 1:
        try:
            decPart = int(splitted[1])
            if decPart != 0:
                res += sep + splitted[1]
            if removeTrailingZeros: res = res.rstrip('0')
        except ValueError:
            # This exception may occur when the float value has an "exp"
            # part, like in this example: 4.345e-05
            res += sep + splitted[1]
    return res

def roundNumber(n, base=5):
    '''Rounds an integer number p_n to an integer value being p_base'''
    return int(base * round(float(n)/base))

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def multicall(o, name, block, *args):
    '''Via multiple inheritance, method named p_name may exist, for instance
       p_o, on its base class from the framework AND on its base class from the
       app. This method may execute both methods, only one of them or does
       nothing if method name p_name is not defined at any level.'''
    # ~~~
    # This is typically used to call framework methods like "onEdit": the method
    # first calls the base method provided by the framework, then the method
    # defined in the app.
    # ~~~
    # If p_block is True, if the framework method returns False or equivalent,
    # the app method is not called.
    # ~~~
    # (block mode only) The method returns the result of the last called method,
    # or True if no method was called at all.
    # ~~~
    r = True if block else None
    bases = o.__class__.__bases__
    for i in (1, 0):
        if hasattr(bases[i], name):
            res = getattr(bases[i], name)(o, *args)
            if block:
                if (i==1) and not res: return
                r = res
            else:
                # If not in "block" mode, don't lose string results
                if isinstance(res, str):
                    if not isinstance(r, str):
                        r = res
                    else:
                        r += res
                elif not isinstance(r, str):
                    r = res
    return r

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def iconParts(icon):
    '''Retrieve the <app> and <name> parts for this p_icon'''
    if '/' in icon:
        base, icon = icon.split('/', 1)
    else:
        base = None
    return icon, base

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
breakpoint = pdb.Pdb(nosigint=True).set_trace
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# -*- coding: utf-8 -*-
# I need a class
class D:
    def getAt1(self):
        return '\n<p>Test1<br /></p>\n'
dummy = D()

titles = '''
<h1>Title 1</h1>
<h2>Title 2</h2>
<h3>Title 3</h3>
<h4>Title 4</h4>
<h5>Title 5</h5>
<h6>Title 6</h6>'''

defaultStyles = '''
<p>Vu que...</p>
<p>Vu que 2...</p>
<!-- A page break after this para -->
<p style="page-break-after: always">Page break after this paragraph.</p>
<p style="margin-left:25px">Attendu que...;</p>
<div style="page-break-before: always">Page break before this one.</div>
'''
